# coding: utf-8

"""
Methods used by BIDS converters
"""
__author__ = "Sabrina Fontanella"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Simona Bottani"
__email__ = "simona.bottani@icm-institute.org"
__status__ = "Completed"


# -- Methods for the clinical data --
# @ToDo:test this function
def create_participants_df(study_name, clinical_spec_path, clinical_data_dir, bids_ids, delete_non_bids_info=True):
    """
    Create the file participants.tsv

    Args:
        study_name: name of the study (Ex. ADNI)
        clinical_spec_path: path to the clinical file
        clinical_data_dir: path to the directory where the clinical data are
        stored
        bids_ids: list of bids ids
        delete_non_bids_info: if True delete all the rows of the subjects that
        are not available in the BIDS dataset

    Returns: a pandas dataframe that contains the participants data
    """
    import pandas as pd
    import os
    from os import path
    import numpy as np
    from clinica.utils.stream import cprint

    fields_bids = ['participant_id']
    prev_location = ''
    index_to_drop = []
    subjects_to_drop = []
    location_name = study_name + ' location'

    # Load the data from the clincal specification file
    participants_specs = pd.read_excel(clinical_spec_path, sheetname='participant.tsv')
    participant_fields_db = participants_specs[study_name]
    field_location = participants_specs[location_name]
    participant_fields_bids = participants_specs['BIDS CLINICA']

    # Extract the list of the available BIDS fields for the dataset
    for i in range(0, len(participant_fields_db)):
        if not pd.isnull(participant_fields_db[i]):
            fields_bids.append(participant_fields_bids[i])

    # Init the dataframe that will be saved in the file participants.tsv
    participant_df = pd.DataFrame(columns=fields_bids)

    for i in range(0, len(participant_fields_db)):
        # If a field not empty is found
        if not pd.isnull(participant_fields_db[i]):
            # Extract the file location of the field and read the value from the file
            tmp = field_location[i].split('/')
            location = tmp[0]
            # If a sheet is available
            if len(tmp) > 1:
                sheet = tmp[1]
            else:
                sheet = ''
            # Check if the file to open for a certain field is the same of the previous field
            if location == prev_location and sheet == prev_sheet:
                pass
            else:
                file_ext = os.path.splitext(location)[1]
                file_to_read_path = path.join(clinical_data_dir, location)

                if file_ext == '.xlsx':
                    file_to_read = pd.read_excel(file_to_read_path, sheetname=sheet)
                elif file_ext == '.csv':
                    file_to_read = pd.read_csv(file_to_read_path)
                prev_location = location
                prev_sheet = sheet

            field_col_values = []
            # For each field in fields_dataset extract all the column values
            for j in range(0, len(file_to_read)):
                # Convert the alternative_id_1 to string if is an integer/float
                value_to_read = file_to_read[participant_fields_db[i]]
                if participant_fields_bids[i] == 'alternative_id_1' and\
                        (value_to_read.dtype == np.float64 or value_to_read.dtype == np.int64):
                    if not pd.isnull(file_to_read.get_value(j, participant_fields_db[i])):
                        value_to_append = str(file_to_read.get_value(j, participant_fields_db[i])).rstrip('.0')
                    else:
                        value_to_append = np.NaN
                else:
                    value_to_append = file_to_read.get_value(j, participant_fields_db[i])
                field_col_values.append(value_to_append)
            # Add the extracted column to the participant_df
            participant_df[participant_fields_bids[i]] = pd.Series(field_col_values)

    if study_name == 'ADNI' or study_name == 'AIBL':
        # ADNImerge contains one row for each visits so there are duplicates
        participant_df = participant_df.drop_duplicates(subset=['alternative_id_1'], keep='first')
    elif study_name == 'OASIS':
        # OASIS provides several MRI for the same session
        participant_df = participant_df[~participant_df.alternative_id_1.str.endswith("_MR2")]
    participant_df.reset_index(inplace=True, drop=True)

    # Adding participant_id column with BIDS ids
    for i in range(0, len(participant_df)):
        if study_name == 'OASIS':
            value = (participant_df['alternative_id_1'][i].split("_"))[1]
        else:
            value = remove_space_and_symbols(participant_df['alternative_id_1'][i])

        bids_id = [s for s in bids_ids if value in s]

        if len(bids_id) == 0:
            index_to_drop.append(i)
            subjects_to_drop.append(value)
        else:
            participant_df.set_value(i, 'participant_id', bids_id[0])

    if len(subjects_to_drop) > 0:
        cprint('The following subjects of ADNIMERGE were not found in your BIDS folder :\n'
               + ', '.join(subjects_to_drop))
    # Delete all the rows of the subjects that are not available in the BIDS dataset
    if delete_non_bids_info:
        participant_df = participant_df.drop(index_to_drop)

    return participant_df


def create_sessions_dict(clinical_data_dir, study_name, clinical_spec_path, bids_ids, name_column_ids, subj_to_remove=[]):
    """

    Extract the information regarding the sessions and store them in a dictionary (session M00 only)

    Args:
        clinical_data_dir: path to the input folder
        study_name: name of the study (Ex: ADNI)
        clinical_spec_path:  path to the clinical file
        bids_ids: list of bids ids
        name_column_ids: name of the column where the subject ids are stored
        subj_to_remove: subjects to remove

    """
    import pandas as pd
    from os import path
    import numpy as np
    import os

    # Load data
    location = study_name + ' location'
    sessions = pd.read_excel(clinical_spec_path, sheetname='sessions.tsv')
    sessions_fields = sessions[study_name]
    field_location = sessions[location]
    sessions_fields_bids = sessions['BIDS CLINICA']
    fields_dataset = []
    fields_bids = []
    sessions_dict = {}

    for i in range(0, len(sessions_fields)):
        if not pd.isnull(sessions_fields[i]):
            fields_bids.append(sessions_fields_bids[i])
            fields_dataset.append(sessions_fields[i])

    sessions_df = pd.DataFrame(columns=fields_bids)

    for i in range(0, len(sessions_fields)):
        # If the i-th field is available
        if not pd.isnull(sessions_fields[i]):
            # Load the file
            tmp = field_location[i].split('/')
            location = tmp[0]
            if len(tmp) > 1:
                sheet = tmp[1]
            else:
                sheet = ''

            file_to_read_path = path.join(clinical_data_dir, location)
            file_ext = os.path.splitext(location)[1]

            if file_ext == '.xlsx':
                file_to_read = pd.read_excel(file_to_read_path, sheetname=sheet)
            elif file_ext == '.csv':
                file_to_read = pd.read_csv(file_to_read_path)

            for r in range(0, len(file_to_read.values)):
                row = file_to_read.iloc[r]
                # Extracts the subject ids columns from the dataframe
                subj_id = row[name_column_ids]
                if hasattr(subj_id, 'dtype'):
                    if subj_id.dtype == np.int64:
                        subj_id = str(subj_id)
                # Removes all the - from
                subj_id_alpha = remove_space_and_symbols(subj_id)
                if study_name == 'OASIS':
                    subj_id_alpha = str(subj_id[0:3] + 'IS' + subj_id[3] + subj_id[5:9])

                # Extract the corresponding BIDS id and create the output file if doesn't exist
                subj_bids = [s for s in bids_ids if subj_id_alpha in s]
                if len(subj_bids) == 0:
                    # If the subject is not an excluded one
                    if subj_id not in subj_to_remove:
                        print(sessions_fields[i] + ' for ' + subj_id + ' not found in the BIDS converted.')
                else:
                    subj_bids = subj_bids[0]
                    sessions_df[sessions_fields_bids[i]] = row[sessions_fields[i]]
                    if subj_bids in sessions_dict:
                        (sessions_dict[subj_bids]['M00']).update({sessions_fields_bids[i]: row[sessions_fields[i]]})
                    else:
                        sessions_dict.update({subj_bids: {
                            'M00': {'session_id': 'ses-M00', sessions_fields_bids[i]: row[sessions_fields[i]]}}})

    return sessions_dict


def create_scans_dict(clinical_data_dir, study_name, clinic_specs_path, bids_ids, name_column_ids):
    """

    Args:
        clinical_data_dir: path to the directory where the clinical data are stored
        study_name: name of the study (Ex ADNI)
        clinic_specs_path: path to the clinical specification file
        bids_ids: list of bids ids
        name_column_ids: name of the column where the subject id is contained

    Returns:

    """
    import pandas as pd
    from os import path
    scans_dict = {}
    prev_file = ''
    prev_sheet = ''

    if study_name not in get_supported_dataset():
        raise Exception('Dataset not supported. Supported datasets are:', get_supported_dataset())

    # Init the dictionary with the subject ids
    for bids_id in bids_ids:
        scans_dict.update({bids_id: {'T1/DWI/fMRI/FMAP': {}, 'FDG': {}}})

    scans_specs = pd.read_excel(clinic_specs_path, sheetname='scans.tsv')
    fields_dataset = []
    fields_location = []
    fields_bids = []
    fields_mod = []

    # Extract the fields available and the corresponding bids name, location and type
    for i in range(0, len(scans_specs[study_name])):
        field = scans_specs[study_name][i]
        if not pd.isnull(field):
            fields_dataset.append(field)
            fields_bids.append(scans_specs['BIDS CLINICA'][i])
            fields_location.append(scans_specs[study_name + ' location'][i])
            fields_mod.append(scans_specs['Modalities related'][i])

    # For each field available extract the original name, extract from the file all the values and fill a data structure
    for i in range(0, len(fields_dataset)):
        # Location is composed by file/sheet
        location = fields_location[i].split('/')
        file_name = location[0]
        if len(location) > 1:
            sheet = location[1]
        else:
            sheet = ''
        # Check if the file to read is already opened
        if file_name == prev_file and sheet == prev_sheet:
            pass
        else:
            file_to_read_path = path.join(clinical_data_dir, file_name)
            if sheet != '':
                file_to_read = pd.read_excel(file_to_read_path, sheetname=sheet)
            else:
                file_to_read = pd.read_excel(file_to_read_path)
            prev_file = file_name
            prev_sheet = sheet

        for bids_id in bids_ids:
            original_id = bids_id.replace('sub-'+study_name, '')
            row_to_extract = file_to_read[file_to_read[name_column_ids] == int(original_id)].index.tolist()
            if len(row_to_extract) > 0:
                row_to_extract = row_to_extract[0]
                # Fill the dictionary with all the information
                (scans_dict[bids_id][fields_mod[i]]).update(
                    {fields_bids[i]: file_to_read.iloc[row_to_extract][fields_dataset[i]]})
            else:
                print(" Scans information for " + bids_id + " not found.")

    return scans_dict


def write_sessions_tsv(bids_dir, sessions_dict):
    """
    Write the content of the function create scans dict in several tsv files following the BIDS specification

    Args:
        bids_dir: path to the bids directory
        sessions_dict: output of the function create_scans_dict

    Returns:

    """
    import os
    import pandas as pd
    from os import path
    from glob import glob

    bids_paths = glob(path.join(bids_dir, 'sub-*'))

    for sp in bids_paths:
        bids_id = sp.split(os.sep)[-1]

        if bids_id in sessions_dict:
            session_df = pd.DataFrame(sessions_dict[bids_id]['M00'], index=['i', ])
            cols = session_df.columns.tolist()
            cols = cols[-1:] + cols[:-1]
            session_df = session_df[cols]
            session_df.to_csv(path.join(sp, bids_id + '_sessions.tsv'), sep='\t', index=False, encoding='utf8')
        else:
            print("No session data available for " + sp)
            session_df = pd.DataFrame(columns=['session_id'])
            session_df['session_id'] = pd.Series('M00')
            session_df.to_csv(path.join(sp, bids_id + '_sessions.tsv'), sep='\t', index=False, encoding='utf8')


def write_scans_tsv(bids_dir, bids_ids, scans_dict):
    """

    Write the scans dict into tsv files

    Args:
        bids_dir:  path to the BIDS directory
        bids_ids: list of bids ids
        scans_dict:  the output of the function create_scans_dict

    """
    import pandas as pd
    from os import path
    import os
    from glob import glob

    for bids_id in bids_ids:
        scans_df = pd.DataFrame()
        bids_id = bids_id.split(os.sep)[-1]
        # Create the file
        tsv_name = bids_id + "_ses-M00_scans.tsv"
        # If the file already exists, remove it
        if os.path.exists(path.join(bids_dir, bids_id, 'ses-M00', tsv_name)):
            os.remove(path.join(bids_dir, bids_id, 'ses-M00', tsv_name))

        mod_available = glob(path.join(bids_dir, bids_id, 'ses-M00', '*'))
        for mod in mod_available:
            mod_name = os.path.basename(mod)
            files = glob(path.join(mod, '*'))
            for file in files:
                file_name = os.path.basename(file)
                if mod_name == "anat" or mod_name == "dwi" or mod_name == "func":
                    f_type = 'T1/DWI/fMRI/FMAP'
                elif mod_name == 'pet':
                    f_type = 'FDG'

                row_to_append = pd.DataFrame(scans_dict[bids_id][f_type], index=[0])
                # Insert the column filename as first value
                row_to_append.insert(0, 'filename', path.join(mod_name, file_name))
                scans_df = scans_df.append(row_to_append)

            scans_df.to_csv(path.join(bids_dir, bids_id, 'ses-M00', tsv_name), sep='\t', index=False, encoding='utf8')


# -- Other methods --
def contain_dicom(folder_path):
    """
     Check if a folder contains DICOM images

    Args:
        folder_path: path to the folder

    Returns: True if dicom files are found inside the folder, False otherwise

    """
    from glob import glob
    from os import path

    dcm_files = glob(path.join(folder_path, '*.dcm'))
    if len(dcm_files) > 0:
        return True

    return False


def get_supported_dataset():
    """
    Return the list of supported datasets

    Returns: a list of supported datasets

    """
    return ['ADNI', 'CLINAD', 'PREVDEMALS', 'INSIGHT', 'OASIS', 'AIBL']


def dcm_to_nii(input_path, output_path, bids_name):
    """
    Convert DICOM to NIFTI trying to use several converters

    Args:
        input_path: path to the input folder with DICOM images
        output_path: path to the output folder
        bids_name: bids name to give to the converterd file

    """
    import os
    from os import path
    import re
    import subprocess
    from glob import glob
    from clinica.utils.stream import cprint
    from colorama import Fore

    if not os.path.exists(output_path):
        os.mkdir(output_path)

    if 'bold' in bids_name:
        # generation of the json file
        cmd = 'dcm2niix -b y -z y -o ' + output_path + ' -f ' + bids_name + ' ' + input_path
    else:
        cmd = 'dcm2niix -b n -z y -o ' + output_path + ' -f ' + bids_name + ' ' + input_path

    subprocess.run(cmd,
                   shell=True,
                   stderr=subprocess.DEVNULL,
                   stdout=subprocess.DEVNULL)

    # If "_t" - the trigger delay time - exists in dcm2niix output filename, we remove it
    exception_t = glob(path.join(output_path, bids_name + '_t[0-9]*'))
    for trigger_time in exception_t:
        res = re.search('_t\d+\.', trigger_time)
        no_trigger_time = trigger_time.replace(trigger_time[res.start(): res.end()], '.')
        os.rename(trigger_time, no_trigger_time)

    # If dcm2niix didn't work use dcm2nii
    if not os.path.exists(path.join(output_path, bids_name + '.nii.gz')):
        cprint('\tConversion with dcm2niix failed, trying with dcm2nii')
        cmd = 'dcm2nii -a n -d n -e n -i y -g n -p n -m n -r n -x n -o ' + output_path + ' ' + input_path
        subprocess.run(cmd,
                       shell=True,
                       stderr=subprocess.DEVNULL,
                       stdout=subprocess.DEVNULL)

    # If the conversion failed with both tools
    if not os.path.exists(path.join(output_path, bids_name + '.nii.gz')):
        cprint(Fore.RED + 'Conversion of the dicom failed for ' + input_path + Fore.RESET)


def get_bids_subjs_list(bids_path):
    """

    Given a BIDS compliant dataset, returns the list of all the subjects available

    Args:/
        bids_path: path to the BIDS folder

    """
    import os
    from os import path

    return [d for d in os.listdir(bids_path) if os.path.isdir(path.join(bids_path, d))]


def get_bids_subjs_paths(bids_path):
    """

    Given a BIDS compliant dataset, returns the list of all paths to the subjects folders

    Args:
        bids_path: path to the BIDS folder
    """

    import os
    from os import path

    return [path.join(bids_path, d) for d in os.listdir(bids_path) if os.path.isdir(path.join(bids_path, d))]


def compute_new_subjects(original_ids, bids_ids):
    """
    Check for new subject to convert

    This function checks for news subjects to convert to the BIDS version i.e. subjects contained in the unorganised
    version that are not available in the bids version.


    Args:
        original_ids: list of all the ids of the unorganized folder.
        bids_ids: list of all the BIDS ids contained inside the bids converted version of the dataset

    Returns: a list containing the original_ids of the subjects that are not available in the bids converted version

    """
    to_return = []
    original_ids = remove_space_and_symbols(original_ids)

    for s in original_ids:
        if not any(s in id for id in bids_ids):
            to_return.append(s)

    return to_return


def remove_space_and_symbols(data):
    '''
    Remove spaces and  - _ from a list (or a single) of strings
    :param data: list of strings or a single string to clean
    :return:
        data: list of strings or a string without space and symbols _ and -
    '''

    import re

    if type(data) is list:
        for i in range(0, len(data)):
            data[i] = re.sub('[-_ ]', '', data[i])
    else:
        data = re.sub('[-_ ]', '', data)

    return data


def get_ext(file_path):

    import os

    root, ext = os.path.splitext(file_path)
    if ext in '.gz':
        file_ext = os.path.splitext(root)[1] + ext
    else:
        file_ext = ext
    return file_ext


def compress_nii(file_path):
    '''
    Compress nii files.

    :param file_path: path to the file to convert
    '''
    from os import remove
    import gzip
    import shutil

    with open(file_path, 'rb') as f_in:
        with gzip.open(file_path + '.gz', 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)
    remove(file_path)


def remove_rescan(list_path):
    """
    Remove all the folders containing the keyword 'rescan' from
    a given list of folders.

    Args:
        list_path (str): the list of the files to analize.

    Returns:
        The list of non rescanned folders.

    """

    import logging

    # Extract all the folders without the substring 'rescan'
    no_resc_lst = [s for s in list_path if 'rescan' not in s]
    if len(no_resc_lst) != len(list_path):
        for r_file in list(set(list_path)-set(no_resc_lst)):
            logging.warning('Rescan found '+r_file+' Ignored.')

    return no_resc_lst


def has_fixed(subj_id, ses, mod, special_list):
    """
    Check if a certain subject has some fixed folder name to convert for a certain modality.

    Args:
        subj_id: the subject to check
        ses: the session
        mod: the modality
        special_list: json file that contains the list of subjects and modality to convert

    Returns:
         False: if the fixed modality to convert is not available for the specified subject
         file_to_convert: name of the folder to choosefor the conversion
    """

    if subj_id in special_list:
        if mod in special_list[subj_id]:
            if ses in special_list[subj_id][mod]:
                file_to_convert = special_list[subj_id][mod]
                return file_to_convert
            else:
                return False
    return False


def choose_correction(dir, to_consider, mod):
    """
    Decides what is the best file type to choose depending on a priority list.

    The selection criteria is based on the input list toConsider that contains
    the files to consider in order of priority.

    Args:
        dir: the directory containing the files.
        to_consider: the list of files to consider in order of priority.
        mod: the modality

    Returns:
        The best file (according to the list to_consider) for the given modality or:
        -1 if if the modality is missing.
        0 if none of the desided corrections is available..

    """

    from os import path
    from glob import glob
    import os

    # Extract all the files available for a certain modality
    correction_list = remove_rescan(glob(path.join(dir, '*' + mod + '*')))
    if len(correction_list) == 0:
        return -1
    if len(correction_list) == 1:
        return correction_list[0].split(os.sep)[-1]
    else:
        for i in range(0, len(to_consider)):
            if any(to_consider[i] in c for c in correction_list):
                return to_consider[i]
        return 0


def get_bids_suff(mod):
    """
    Returns the BIDS suffix for a certain modality.

    Args:
        mod: modality.
    Returns:
        The suffix used in the BIDS standard for a certain modality.
    """

    bids_suff = {
        'T1': '_T1w',
        'T2': '_T2w',
        'Flair': '_FLAIR',
        'SingleMapPh': '_phasediff',
        'MultiMapPh': '_phase',
        'Map': '_magnitude',
        'fMRI': '_bold',
        'dwi': '_dwi',
        'pet': '_pet'
    }

    return bids_suff[mod]


def create_path_corr_file(out_dir):
    import os
    from os import path

    path_corr = path.join(out_dir, 'conversion_info', 'filename_correspondance.tsv')
    if os.path.exists(path_corr):
        os.remove(path_corr)

    f = open(path_corr, 'a')
    return f


def convert_T1(t1_path, output_path, t1_bids_name):
    """
    Convert into the BIDS specification a T1 image.

    Args:
        t1_path: the path of the T1 images to convert.
        output_path: output folder path.
        t1_bids_name: name to give to the file.

    """
    from os import path
    from shutil import copy
    import os

    bids_name = path.join(output_path, t1_bids_name + get_bids_suff('T1'))

    if contain_dicom(t1_path):
        print('DICOM found for t1 in ' + t1_path)
        dcm_to_nii(t1_path, output_path, t1_bids_name + get_bids_suff('T1'))
    else:
        if not os.path.exists(output_path):
            os.mkdir(output_path)
        file_ext = get_ext(t1_path)
        bids_file = bids_name + file_ext
        print(bids_file)
        copy(t1_path, bids_file)
        # If  the original image is not compress, compress it
        if file_ext == '.nii':
            compress_nii(path.join(bids_file))


def convert_pet(folder_input, folder_output, pet_name, bids_name, task_name, acquisition=''):
    """
    Convert PET to BIDS

    Args:
        folder_input: path to the input folder
        folder_output: path to the BIDS folder
        pet_name: original name
        bids_name: bids name
        task_name: task name
        acquisition: acquisition name

    """

    from os import path
    from shutil import copy
    import os

    if acquisition != '':
        pet_bids_name = bids_name+'_task-'+task_name+'_acq-'+acquisition
    else:
        pet_bids_name = bids_name+'_task-'+task_name

    if contain_dicom(folder_input):
        print('DICOM found for PET')
        dcm_to_nii(folder_input, folder_output, pet_bids_name, 'pet')
    else:
        if not os.path.exists(folder_output):
            os.mkdir(folder_output)

        # If a prefixed pet is chosen for the conversion use it, otherwise extracts the pet file available into the folder
        if pet_name != '':
            pet_path = path.join(folder_input, pet_name)
        else:
            print('WARNING: feature to be implemented')

        file_ext = get_ext(pet_path)
        copy(pet_path, path.join(folder_output, pet_bids_name + get_bids_suff('pet') + file_ext))
        # If  the original image is not compress, compress it
        if file_ext == '.nii':
            compress_nii(path.join(folder_output, pet_bids_name + get_bids_suff('pet') + file_ext))


def convert_fieldmap(folder_input, folder_output, name, fixed_file=[False, False]):
    """
    Extract and convert into the BIDS specification fieldmap data.

    Args:
        folder_input: folder containing the fieldmap data.
        folder_output: folder where store the converted files.
        name: name of converted file.
        files_to_skip:

    Returns:
         -1 if the modality is not available.
         0 if the magnitude or the phase is missing (information incomplete).
    """

    from os import path
    from glob import glob
    from shutil import copy
    import nibabel as nib
    import os

    mag_missing = map_ph_missing = False

    # Check if there is a map or mapPh fixed to convert
    if fixed_file[0] is False and fixed_file[1] is False:
        map = remove_rescan(glob(path.join(folder_input, "*MAP_*", '*.nii.gz')))
        map_ph = remove_rescan(glob(path.join(folder_input, "*MAPph_*", '*.nii.gz')))
    elif fixed_file[0] is False and fixed_file[1] is not False:
        map = remove_rescan(glob(path.join(folder_input, "*MAP_*", '*.nii.gz')))
        map_ph = glob(path.join(folder_input, fixed_file[1], '*.nii.gz'))
    elif fixed_file[0] is not False and fixed_file[1] is False:
        map = glob(path.join(folder_input, fixed_file[0], '*.nii.gz'))
        map_ph = remove_rescan(glob(path.join(folder_input, "*MAPph_*", '*.nii.gz')))
    else:
        map = glob(path.join(folder_input, fixed_file[0], '*.nii.gz'))
        map_ph = glob(path.join(folder_input, fixed_file[1], '*.nii.gz'))

    files_to_skip = []
    if len(map) == 0:
        mag_missing = True
    if len(map_ph) == 0:
        map_ph_missing = True
    # If the information regarding the Fieldmap data are complete
    if len(map) > 0 and len(map_ph) > 0:
        map_ph_name = map_ph[0].split(os.sep)[-1]
        map_name = map[0].split(os.sep)[-1]

        # toSolve: there are some files that produce an error when loaded with Nibabel
        if (map_ph_name not in files_to_skip) and (map_name not in files_to_skip):
            # Open the files with Nibabel
            map_nib = nib.load(map[0])
            map_ph_nib = nib.load(map_ph[0])
            dim_map = (map_nib.header['dim'])[4]
            dim_map_ph = (map_ph_nib.header['dim'])[4]
            os.mkdir(path.join(folder_output))
            # Case 1: one phase difference image and at least one magnitude image
            if dim_map_ph == 1 and dim_map > 0:
                copy(map_ph[0], path.join(folder_output, name + get_bids_suff('SingleMapPh') + '.nii.gz'))
                os.system('fslsplit ' + map[0] + ' ' + path.join(folder_output, name + get_bids_suff('Map')))
                mag_list = glob(path.join(folder_output, name+get_bids_suff('Map')+'*'))

                for i in range(0, len(mag_list)):
                    old_mag_name = mag_list[i].split(os.sep)[-1]
                    # Remove the extension and the number sequence of fslsplit
                    new_mag_name = old_mag_name[:-11] + str(i+1)
                    os.rename(mag_list[i], path.join(folder_output, new_mag_name + ".nii.gz"))
            # Case 2: two phase images and two magnitude images'
            elif dim_map_ph == 2 and dim_map == 2:
                bids_name_ph = name + get_bids_suff('MultiMapPh')
                bids_name_mag = name + get_bids_suff('Map')
                os.system('fslsplit ' + map_ph[0] + ' ' + path.join(folder_output, bids_name_ph))
                os.system('fslsplit ' + map[0] + ' ' + path.join(folder_output, bids_name_mag))
                # Fslsplit produces files with a name not comply with BIDS specification, a rename is needed
                os.rename(path.join(folder_output, bids_name_mag + '0000.nii.gz'),
                          path.join(folder_output, bids_name_mag+'1.nii.gz'))
                os.rename(path.join(folder_output, bids_name_mag + '0001.nii.gz'),
                          path.join(folder_output, bids_name_mag + '2.nii.gz'))
                os.rename(path.join(folder_output, bids_name_ph + '0000.nii.gz'),
                          path.join(folder_output, bids_name_ph + '1.nii.gz'))
                os.rename(path.join(folder_output, bids_name_ph + '0001.nii.gz'),
                          path.join(folder_output, bids_name_ph + '2.nii.gz'))
    # The modalities is missing or incomplete
    else:
        if mag_missing and map_ph_missing:
            return -1
        else:
            return 0


def convert_flair(folder_input, folder_output, name, fixed_file=False):
    """
    Extracts and converts T2Flair data.

    Args:
        folder_input: folder containing T2Flair.
        folder_output: output folder path.
        name: name to give to the output file.

    Returns:
        -1 if no T2FLAIR is found in the input folder.
    """

    from os import path
    import logging
    from glob import glob
    from shutil import copy
    import os

    # If a given T2FLAIR is given for the conversion use it
    if fixed_file:
        fixed_flair_path = glob(path.join(folder_input, fixed_file, '*'))[0]
        copy(fixed_flair_path, path.join(folder_output, name + (get_bids_suff('Flair')) + '.nii.gz'))
    else:
        list_path = glob(path.join(folder_input, '*T2FLAIR*'))
        flair_lst = remove_rescan(list_path)
        if len(flair_lst) == 1:
            if not os.path.exists(folder_output):
                os.mkdir(folder_output)
            flair_path = glob(path.join(flair_lst[0], '*.nii.gz*'))[0]
            copy(flair_path, path.join(folder_output, name + (get_bids_suff('Flair')) + '.nii.gz'))
        elif len(flair_lst) == 0:
            return -1
        elif len(flair_lst) > 1:
            logging.warning('Multiple FLAIR found, computation aborted.')
            raise('Aborted')


def convert_fmri(folder_input, folder_output, name, fixed_fmri=False, task_name='rest'):
    """
    Extracts and converts into the BIDS specification fmri data.

    Args:
        folder_input: the folder containing the fmri to convert.
        folder_output: the output folder.
        name:

    Returns:
        -1 if no fMRI file is found in the input folder.
    """

    from os import path
    import logging
    from glob import glob
    from shutil import copy
    import os
    from clinica.utils.stream import cprint
    from colorama import Fore

    bids_name = name + '_task-' + task_name + get_bids_suff('fMRI')

    if contain_dicom(folder_input):
        dcm_to_nii(folder_input, folder_output, bids_name)
    else:
        if fixed_fmri is not False:
            fmri_lst = glob(path.join(folder_input, fixed_fmri))
        else:
            fmri_lst = remove_rescan(glob(path.join(folder_input, '*fMRI*')))
        if len(fmri_lst) > 0:
            if not os.path.exists(folder_output):
                os.mkdir(folder_output)
            fmri_file_path = glob(path.join(fmri_lst[0], '*.nii*'))[0]
            file_ext = get_ext(fmri_file_path)
            copy(fmri_file_path, path.join(folder_output, bids_name + file_ext))
            if file_ext == '.nii':
                cprint(Fore.YELLOW + 'Non compressed file found: ' + fmri_file_path + Fore.RESET)
                compress_nii(path.join(folder_output, bids_name + file_ext))
        else:
            logging.info('Non fMRI found for ' + folder_input)
            return -1


def merge_DTI(folder_input, folder_output, name, fixed_dti_list=False):
    """
    Merge all the DTI files of a given subject.

    For the merging only DTI folders containing all .nii.gz, .bval and .bvec are considered,
    otherwise the folder is ignored.

    Args:
        folder_input (str) : the folder containing all the DTI to merge.
        folder_output (str) : the folder where store the merged file.
        name (str): name to give to the merged file.

    Returns:
        -1 if the input folder doesn't contain any DTI folder.
        The list of incomplete DTI folders if there is some folders without bvec/bval/nii
    """

    from os import path
    from glob import glob
    import fileinput
    import os

    img = []
    bval = []
    bvec = []
    dti_list = []
    # The bvec output file has three rows that are the concatenation of all the merged bvec files
    lines_out_bvec = ['', '', '']
    # The bval output file has only one row that is the concatenation of all the merged bval files
    lines_out_bval = ['']

    if fixed_dti_list is not False:
        for dti in fixed_dti_list:
            dti_list.append(path.join(folder_input, dti))
    else:
        dti_list = remove_rescan(glob(path.join(folder_input, '*DTI*')))
    incomp_folders = []
    nr_dti = len(dti_list)
    if nr_dti == 0:
        return -1
    else:
        if not os.path.exists(folder_output):
            os.mkdir(folder_output)
        for folder in dti_list:
            if len(glob(path.join(folder, '*.bval'))) != 0 and len(glob(path.join(folder, '*.bvec'))) != 0:
                img.append(glob(path.join(folder, '*.nii*'))[0])
                bval.append(glob(path.join(folder, '*.bval'))[0])
                bvec.append(glob(path.join(folder, '*.bvec'))[0])
            else:
                incomp_folders.append(folder)

        # if it has been found at least a DTI folder complete with bvec, bval and nii.gz
        if len(img) > 0:
            file_suff = get_bids_suff('dwi')
            fin = fileinput.input(bval)
            # merge all the .nii.gz file with fslmerge
            os.system('fslmerge -t ' + path.join(folder_output, name + file_suff + '.nii.gz') + ' ' + " ".join(img))
            # merge all the .bval files
            fout = open(path.join(folder_output, name+file_suff+'.bval'), 'w')
            # Concatenate bval files
            for line in fin:
                if fileinput.isfirstline():
                    line_no = 0
                lines_out_bval[line_no] = lines_out_bval[line_no]+" "+line.rstrip()
                line_no += 1
            for i in range(0, len(lines_out_bval)):
                lines_out_bval[i] = lines_out_bval[i].lstrip()

                fout.write(lines_out_bval[i]+"\n")

            # Concatenate bvec files
            fin = fileinput.input(bvec)
            fout = open(path.join(folder_output, name + file_suff + '.bvec'), 'w')
            for line in fin:
                if fileinput.isfirstline():
                    line_no = 0

                lines_out_bvec[line_no] = lines_out_bvec[line_no]+" "+line.rstrip()
                line_no += 1
            for i in range(0, len(lines_out_bvec)):
                lines_out_bvec[i] = lines_out_bvec[i].lstrip()
                fout.write(lines_out_bvec[i] + "\n")

        if len(incomp_folders) > 0:
            return incomp_folders


def concatenate_bvec_bval(files_list, output_file, type):
    """
    Concatenate bvec and baval files

    Args:
        files_list: list of file
        output_file: path to the output file
        type: bvec of bval

    """
    import fileinput
    if type == 'bval':
        lines_out = ['']
    else:
        lines_out = ['', '', '']

    for line in files_list:
        if fileinput.isfirstline():
            line_no = 0
        lines_out[line_no] = lines_out[line_no] + " " + line.rstrip()
        line_no += 1
    for i in range(0, len(lines_out)):
        lines_out[i] = lines_out[i].lstrip()

        output_file.write(lines_out[i] + "\n")


def merge_noddi_dti(folder_input, folder_output, name):
    """
       Merge NODDI dti CATI organised following these rules:

       - DTI1, DTI3 and DTI5 are always Anterior to Posterior(AP/j-)
       - DTI2, DTI4, DTI6 are always Posterior to Anterior(PA/j)

    Args:
        folder_input: path to the folder where are the DTI
        folder_output: path to the BIDS folder
        name: BIDS id of the file
    """
    from glob import glob
    from os import path
    import fileinput
    import os

    dti_pa_nii = []
    dti_ap_nii = []
    dti_pa_bvec = []
    dti_ap_bvec = []
    dti_pa_bval = []
    dti_ap_bval = []

    dti_list = remove_rescan(glob(path.join(folder_input, '*DTI*')))

    if len(dti_list) != 6:
        raise Exception('Number of DTI found different from 6')

    if not os.path.exists(folder_output):
        os.mkdir(folder_output)

    out_file_name_pa = path.join(folder_output, name + '_seq-mshellPA' + get_bids_suff('dwi'))
    out_file_name_ap = path.join(folder_output, name + '_seq-mshellAP' + get_bids_suff('dwi'))

    dti_ap_paths = glob(path.join(folder_input, '*DTI[1,3,5]*'))
    dti_pa_paths = glob(path.join(folder_input, '*DTI[2,4,6]*'))

    for f in dti_pa_paths:
        dti_pa_nii.append(glob(path.join(f, '*.nii*'))[0])
        dti_pa_bvec.append(glob(path.join(f, '*.bvec'))[0])
        dti_pa_bval.append(glob(path.join(f, '*.bval'))[0])

    for f in dti_ap_paths:
        dti_ap_nii.append(glob(path.join(f, '*.nii*'))[0])
        dti_ap_bvec.append(glob(path.join(f, '*.bvec'))[0])
        dti_ap_bval.append(glob(path.join(f, '*.bval'))[0])

    # Merge all the nii files
    os.system('fslmerge -t ' + (out_file_name_pa + '.nii.gz') + ' ' + " ".join(dti_pa_nii))
    os.system('fslmerge -t ' + (out_file_name_ap + '.nii.gz') + ' ' + " ".join(dti_ap_nii))

    # Concatenate all the bvec/bval files
    concatenate_bvec_bval(fileinput.input(dti_pa_bval), open(path.join(out_file_name_pa + '.bval'), 'w'), 'bval')
    concatenate_bvec_bval(fileinput.input(dti_pa_bvec), open(path.join(out_file_name_pa + '.bvec'), 'w'), 'bvec')
    concatenate_bvec_bval(fileinput.input(dti_ap_bval), open(path.join(out_file_name_ap + '.bval'), 'w'), 'bval')
    concatenate_bvec_bval(fileinput.input(dti_ap_bvec), open(path.join(out_file_name_ap + '.bvec'), 'w'), 'bvec')



# coding: utf-8

"""
 Module for converting FDG PET of ADNI
"""

__author__ = "Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Sabrina Fontanella"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


def convert_adni_fdg_pet(source_dir, csv_dir, dest_dir, subjs_list=None):
    """
    Convert FDG PET images of ADNI into BIDS format

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    """

    import pandas as pd
    from os import path
    from clinica.utils.stream import cprint
    from clinica.iotools.converters.adni_to_bids.adni_utils import t1_pet_paths_to_bids
    from colorama import Fore

    if subjs_list is None:
        adni_merge_path = path.join(csv_dir, 'ADNIMERGE.csv')
        adni_merge = pd.read_csv(adni_merge_path, sep=',', low_memory=False)
        subjs_list = list(adni_merge.PTID.unique())

    cprint('Calculating paths of FDG PET images. Output will be stored in %s.' % path.join(dest_dir, 'conversion_info'))
    images = compute_fdg_pet_paths(source_dir, csv_dir, dest_dir, subjs_list)
    cprint('Paths of FDG PET images found. Exporting images into BIDS ...')
    t1_pet_paths_to_bids(images, dest_dir, 'fdg')
    cprint(Fore.GREEN + 'FDG PET conversion done.' + Fore.RESET)


def compute_fdg_pet_paths(source_dir, csv_dir, dest_dir, subjs_list):
    """
    Compute the paths to the FDG PET images and store them in a tsv file

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    Returns:
        images: a dataframe with all the paths to the PET images that will be converted into BIDS

    """

    import pandas as pd
    import os
    import operator
    from os import path
    from functools import reduce
    from clinica.iotools.converters.adni_to_bids.adni_utils import replace_sequence_chars, find_image_path
    from clinica.utils.stream import cprint

    pet_fdg_col = ['Phase', 'Subject_ID', 'VISCODE', 'Visit', 'Sequence', 'Scan_Date', 'Study_ID',
                   'Series_ID', 'Image_ID', 'Original']
    pet_fdg_df = pd.DataFrame(columns=pet_fdg_col)

    # Loading needed .csv files
    petqc = pd.read_csv(path.join(csv_dir, 'PETQC.csv'), sep=',', low_memory=False)
    petqc3 = pd.read_csv(path.join(csv_dir, 'PETC3.csv'), sep=',', low_memory=False)
    pet_meta_list = pd.read_csv(path.join(csv_dir, 'PET_META_LIST.csv'), sep=',', low_memory=False)

    for subj in subjs_list:

        # PET images metadata for subject
        subject_pet_meta = pet_meta_list[pet_meta_list['Subject'] == subj]

        if subject_pet_meta.shape[0] < 1:
            # TODO Log somewhere subjects without FDG-PET images metadata
            continue

        # QC for FDG PET images for ADNI 1, GO and 2
        pet_qc_1go2_subj = petqc[(petqc.PASS == 1) & (petqc.RID == int(subj[-4:]))]

        # QC for FDG PET images for ADNI 3
        pet_qc3_subj = petqc3[(petqc3.SCANQLTY == 1) & (petqc3.RID == int(subj[-4:]))]
        pet_qc3_subj.insert(0, 'EXAMDATE', pet_qc3_subj.SCANDATE.to_list())

        # Concatenating visits in both QC files
        pet_qc_subj = pd.concat([pet_qc_1go2_subj, pet_qc3_subj], axis=0, ignore_index=True, sort=False)

        for visit in list(pet_qc_subj.VISCODE2.unique()):
            pet_qc_visit = pet_qc_subj[pet_qc_subj.VISCODE2 == visit]

            # If there are several scans for a timepoint we keep image acquired last (higher LONIUID)
            pet_qc_visit = pet_qc_visit.sort_values("LONIUID", ascending=False)

            qc_visit = pet_qc_visit.iloc[0]

            # Corresponding LONI image ID for original scan in PET Meta List
            int_image_id = int(qc_visit.LONIUID[1:])

            original_pet_meta = subject_pet_meta[(subject_pet_meta['Orig/Proc'] == 'Original')
                                                 & (subject_pet_meta['Image ID'] == int_image_id)
                                                 & (subject_pet_meta.Sequence.map(lambda s: (s.lower().find('early')
                                                                                             < 0)))]
            # If no corresponding FDG PET metadata for an original image,
            # take scan at the same date containing FDG in sequence name
            if original_pet_meta.shape[0] < 1:
                original_pet_meta = subject_pet_meta[(subject_pet_meta['Orig/Proc'] == 'Original')
                                                     & (subject_pet_meta.Sequence.map(lambda x: (x.lower().find('fdg')
                                                                                                 > -1)))
                                                     & (subject_pet_meta['Scan Date'] == qc_visit.EXAMDATE)]

                if original_pet_meta.shape[0] < 1:
                    # TODO Log somewhere QC visits without image metadata
                    cprint('No FDG-PET images metadata for subject - ' + subj + ' for visit ' + qc_visit.VISCODE2)
                    continue

            original_image = original_pet_meta.iloc[0]

            # Co-registered and Averaged image with the same Series ID of the original image
            averaged_pet_meta = subject_pet_meta[(subject_pet_meta['Sequence'] == 'Co-registered, Averaged')
                                                 & (subject_pet_meta['Series ID'] == original_image['Series ID'])]

            # If an explicit Co-registered, Averaged image does not exist,
            # the original image is already in that preprocessing stage.

            if averaged_pet_meta.shape[0] < 1:
                sel_image = original_image
                original = True
            else:
                sel_image = averaged_pet_meta.iloc[0]
                original = False

            visit = sel_image.Visit
            sequence = replace_sequence_chars(sel_image.Sequence)
            date = sel_image['Scan Date']
            study_id = sel_image['Study ID']
            series_id = sel_image['Series ID']
            image_id = sel_image['Image ID']

            row_to_append = pd.DataFrame(
                [[qc_visit.Phase, subj, qc_visit.VISCODE2, str(visit), sequence, date, str(study_id), str(series_id),
                  str(image_id), original]],
                columns=pet_fdg_col)
            pet_fdg_df = pet_fdg_df.append(row_to_append, ignore_index=True)

    # Exceptions
    # ==========
    conversion_errors = [  # NONAME.nii
                         ('031_S_0294', 'bl'),
                         ('037_S_1421', 'm36'),
                         ('037_S_1078', 'm36'),

                         # Empty folders
                         ('941_S_1195', 'm48'),
                         ('005_S_0223', 'm12')]

    error_indices = []
    for conv_error in conversion_errors:
        error_indices.append((pet_fdg_df.Subject_ID == conv_error[0])
                             & (pet_fdg_df.VISCODE == conv_error[1]))

    indices_to_remove = pet_fdg_df.index[reduce(operator.or_, error_indices, False)]
    pet_fdg_df.drop(indices_to_remove, inplace=True)

    images = find_image_path(pet_fdg_df, source_dir, 'FDG', 'I', 'Image_ID')

    fdg_csv_path = path.join(dest_dir, 'conversion_info')
    if not os.path.exists(fdg_csv_path):
        os.mkdir(fdg_csv_path)
    images.to_csv(path.join(fdg_csv_path, 'fdg_pet_paths.tsv'), sep='\t', index=False)

    return images


def check_exceptions(bids_dir):
    """
    Function for verifying that proper images are in bids folder
    """

    from os import path
    import pandas as pd
    from glob import glob

    fdg_paths = pd.read_csv(path.join(bids_dir, 'conversion_info', 'fdg_pet_paths.tsv'), sep='\t')
    fdg_paths = fdg_paths[fdg_paths.Path.notnull()]
    fdg_paths['BIDS_SubjID'] = ['sub-ADNI' + s.replace('_', '') for s in fdg_paths.Subject_ID.to_list()]
    fdg_paths['BIDS_Session'] = ['ses-' + s.replace('bl', 'm00').upper() for s in fdg_paths.VISCODE.to_list()]

    count = 0
    for r in fdg_paths.iterrows():
        image = r[1]
        image_dir = path.join(bids_dir, image.BIDS_SubjID, image.BIDS_Session, 'pet')
        image_pattern = path.join(image_dir, '%s_%s_*' % (image.BIDS_SubjID, image.BIDS_Session))
        files_list = glob(image_pattern)
        if not files_list:
            print("No images for subject %s in session %s" % (image.BIDS_SubjID, image.BIDS_Session))
            count += 1

        elif len(files_list) > 1:
            print("Too many images for subject %s in session %s" % (image.BIDS_SubjID, image.BIDS_Session))
            print(files_list)

    print(count)



# coding: utf-8

"""
 Module for converting FLAIR of ADNI
"""

__author__ = "Simona Bottani and Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


def convert_adni_flair(source_dir, csv_dir, dest_dir, subjs_list=None):
    """
    Convert FLAIR images of ADNI into BIDS format

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    """

    import pandas as pd
    from os import path
    from clinica.utils.stream import cprint
    from colorama import Fore

    if subjs_list is None:
        adni_merge_path = path.join(csv_dir, 'ADNIMERGE.csv')
        adni_merge = pd.read_csv(adni_merge_path, sep=',', low_memory=False)
        subjs_list = list(adni_merge.PTID.unique())

    cprint('Calculating paths of FLAIR images. Output will be stored in %s.' % path.join(dest_dir, 'conversion_info'))
    images = compute_flair_paths(source_dir, csv_dir, dest_dir, subjs_list)
    cprint('Paths of FLAIR images found. Exporting images into BIDS ...')
    flair_paths_to_bids(images, dest_dir)
    cprint(Fore.GREEN + 'FLAIR conversion done.' + Fore.RESET)


def compute_flair_paths(source_dir, csv_dir, dest_dir, subjs_list):
    """
    Compute the paths to the FLAIR images and store them in a tsv file

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    Returns:
        images: a dataframe with all the paths to the FLAIR images that will be converted into BIDS

    """

    import operator
    from os import path, mkdir
    from functools import reduce
    import pandas as pd

    from clinica.iotools.converters.adni_to_bids.adni_utils import find_image_path, visits_to_timepoints_mrilist

    flair_col_df = ['Subject_ID', 'VISCODE', 'Visit', 'Sequence', 'Scan_Date',
                    'Study_ID', 'Series_ID', 'Image_ID', 'Field_Strength', 'Scanner']
    flair_df = pd.DataFrame(columns=flair_col_df)

    # Loading needed .csv files
    adni_merge = pd.read_csv(path.join(csv_dir, 'ADNIMERGE.csv'), sep=',', low_memory=False)

    mayo_mri_qc = pd.read_csv(path.join(csv_dir, 'MAYOADIRL_MRI_IMAGEQC_12_08_15.csv'), sep=',', low_memory=False)
    mayo_mri_qc = mayo_mri_qc[mayo_mri_qc.series_type == 'AFL']

    mri_list = pd.read_csv(path.join(csv_dir, 'MRILIST.csv'), sep=',', low_memory=False)

    # Selecting FLAIR DTI images that are not TODO images
    mri_list = mri_list[mri_list.SEQUENCE.map(lambda x: x.lower().find('flair') > -1)]
    unwanted_sequences = ['_MPR_']
    mri_list = mri_list[mri_list.SEQUENCE.map(lambda x: not any(subs in x for subs in unwanted_sequences))]

    for subj in subjs_list:

        # Filter ADNIMERGE, MRI_LIST and QC for only one subject and sort the rows/visits by examination date
        adnimerge_subj = adni_merge[adni_merge.PTID == subj]
        adnimerge_subj = adnimerge_subj.sort_values('EXAMDATE')

        mri_list_subj = mri_list[mri_list.SUBJECT == subj]
        mri_list_subj = mri_list_subj.sort_values('SCANDATE')

        mayo_mri_qc_subj = mayo_mri_qc[mayo_mri_qc.RID == int(subj[-4:])]

        # Obtain corresponding timepoints for the subject visits
        visits = visits_to_timepoints_mrilist(subj, mri_list_subj, adnimerge_subj, 'FLAIR')

        for visit_info in visits.keys():
            timepoint = visit_info[0]
            visit_str = visits[visit_info]

            visit_mri_list = mri_list_subj[mri_list_subj.VISIT == visit_str]
            flair = flair_image(subj, timepoint, visits[visit_info], visit_mri_list, mayo_mri_qc_subj)

            if flair is not None:
                row_to_append = pd.DataFrame(flair, index=['i', ])
                # TODO Replace iteratively appending by pandas.concat
                flair_df = flair_df.append(row_to_append, ignore_index=True, sort=False)

    # Exceptions
    # ==========
    conversion_errors = [  # Eq_1 images
                         ('141_S_0767', 'm84'),
                         ('067_S_5205', 'bl'),
                         ('127_S_4928', 'm24'),
                         ('024_S_4674', 'm06'),
                         ('123_S_2363', 'm24'),
                         ('053_S_4578', 'm48'),
                         ('128_S_4586', 'm48'),
                         ('053_S_4813', 'm48'),
                         ('053_S_5272', 'm24'),

    ]

    error_indices = []
    for conv_error in conversion_errors:
        error_indices.append((flair_df.Subject_ID == conv_error[0])
                             & (flair_df.VISCODE == conv_error[1]))
    if error_indices:
        indices_to_remove = flair_df.index[reduce(operator.or_, error_indices, False)]
        flair_df.drop(indices_to_remove, inplace=True)

    # Checking for images paths in filesystem
    images = find_image_path(flair_df, source_dir, 'FLAIR', 'S', 'Series_ID')

    flair_tsv_path = path.join(dest_dir, 'conversion_info')
    if not path.exists(flair_tsv_path):
        mkdir(flair_tsv_path)
    images.to_csv(path.join(flair_tsv_path, 'flair_paths.tsv'), sep='\t', index=False)

    return images


def flair_paths_to_bids(images, dest_dir, mod_to_update=False):
    """
    Convert FLAIR images

    Args:
        images: dataframe returned by the method compute_flair_paths
        dest_dir: path to the destination directory
        mod_to_update: if is true and an image is already existing it will overwrite the old version

    """
    from multiprocessing import cpu_count, Pool, Value
    from functools import partial

    counter = None

    def init(args):
        ''' store the counter for later use '''
        global counter
        counter = args

    subjs_list = [sub for sub in images['Subject_ID'].unique()
                  if sub == sub and 'S' in sub]

    counter = Value('i', 0)
    partial_generate_subject_files = partial(generate_subject_files,
                                             images=images,
                                             dest_dir=dest_dir,
                                             mod_to_update=mod_to_update)
    poolrunner = Pool(cpu_count(), initializer=init, initargs=(counter,))
    poolrunner.map(partial_generate_subject_files, subjs_list)
    del counter


def flair_image(subject_id, timepoint, visit_str, visit_mri_list, mri_qc_subj):
    """

    Args:
        subject_id:
        timepoint:
        visit_str:
        visit_mri_list:
        mri_qc_subj:

    Returns:

    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import replace_sequence_chars, select_image_qc

    sel_image = select_image_qc(list(visit_mri_list.IMAGEUID), mri_qc_subj)
    if sel_image is None:
        return None

    sel_scan = visit_mri_list[visit_mri_list.IMAGEUID == sel_image].iloc[0]

    image_dict = {'Subject_ID': subject_id,
                  'VISCODE': timepoint,
                  'Visit': visit_str,
                  'Sequence': replace_sequence_chars(sel_scan.SEQUENCE),
                  'Scan_Date': sel_scan['SCANDATE'],
                  'Study_ID': str(int(sel_scan.STUDYID)),
                  'Series_ID': str(int(sel_scan.SERIESID)),
                  'Image_ID': str(int(sel_scan.IMAGEUID)),
                  'Field_Strength': sel_scan.MAGSTRENGTH}

    return image_dict


def generate_subject_files(subj, images, dest_dir, mod_to_update):
    import clinica.iotools.bids_utils as bids
    import clinica.iotools.converters.adni_to_bids.adni_utils as adni_utils
    from clinica.utils.stream import cprint
    import subprocess
    import os
    import shutil
    from os import path
    from glob import glob

    alpha_id = bids.remove_space_and_symbols(subj)
    bids_id = 'sub-ADNI' + alpha_id
    # Extract the list of sessions available from the flair paths files, removing the duplicates
    sess_list = images[(images['Subject_ID'] == subj)]['VISCODE'].unique()

    if not os.path.exists(path.join(dest_dir, bids_id)):
        os.mkdir(path.join(dest_dir, bids_id))

    # For each session available, create the folder if doesn't exist and convert the files
    for ses in sess_list:
        with counter.get_lock():
            counter.value += 1
        cprint('[FLAIR] Processing subject ' + str(subj)
               + ' - session ' + ses + ', ' + str(counter.value)
               + ' / ' + str(len(images)))
        ses_bids = adni_utils.viscode_to_session(ses)
        bids_ses_id = 'ses-' + ses_bids
        bids_file_name = bids_id + '_ses-' + ses_bids
        ses_path = path.join(dest_dir, bids_id, bids_ses_id)

        if mod_to_update:
            if os.path.exists(path.join(ses_path, 'FLAIR')):
                shutil.rmtree(path.join(ses_path, 'FLAIR'))

        if not os.path.exists(ses_path):
            os.mkdir(ses_path)

        flair_info = images[
            (images['Subject_ID'] == subj) & (images['VISCODE'] == ses)]

        # For the same subject, same session there could be multiple flair with different acq label
        for j in range(len(flair_info)):
            flair_subj = flair_info.iloc[j]
            if type(flair_subj['Path']) != float and flair_subj['Path'] != '':
                if not os.path.exists(path.join(ses_path, 'FLAIR')):
                    os.mkdir(path.join(ses_path, 'FLAIR'))
                flair_path = flair_subj['Path']
                bids_name = bids_file_name + '_FLAIR'
                bids_dest_dir = path.join(ses_path, 'FLAIR')
                image_id = flair_subj.Image_ID

                # If the original image is a DICOM, check if contains two DICOM
                # inside the same folder
                if flair_subj.Is_Dicom:
                    flair_path = adni_utils.check_two_dcm_folder(flair_path,
                                                                 bids_dest_dir,
                                                                 image_id)

                if not os.path.exists(bids_dest_dir):
                    os.mkdir(dest_dir)
                command = 'dcm2niix -b y -z y -o ' + bids_dest_dir + ' -f ' + bids_name + ' ' + flair_path
                subprocess.run(command,
                               shell=True,
                               stderr=subprocess.DEVNULL,
                               stdout=subprocess.DEVNULL)

                # If dcm2niix didn't work use dcm2nii
                if not os.path.exists(path.join(bids_dest_dir, bids_name + '.nii.gz')):
                    cprint('\tConversion with dcm2niix failed, trying with dcm2nii')

                    # Find all the files eventually created by dcm2niix and remove them
                    flair_dcm2niix = glob(
                        path.join(bids_dest_dir, bids_name + '*'))
                    for d in flair_dcm2niix:
                        os.remove(d)

                    command = 'dcm2nii -a n -d n -e n -i y -g y -p n -m n -r n -x n -o ' + bids_dest_dir + ' ' + flair_path
                    subprocess.run(command,
                                   shell=True,
                                   stdout=subprocess.DEVNULL,
                                   stderr=subprocess.DEVNULL)
                    nii_file = path.join(bids_dest_dir,
                                         subj.replace('_', '') + '.nii.gz')
                    if os.path.exists(nii_file):
                        os.rename(nii_file, path.join(bids_dest_dir,
                                                      bids_name + '.nii.gz'))
                    else:
                        cprint('WARNING: CONVERSION FAILED...')


def check_exceptions(bids_dir):
    from os import path
    import pandas as pd
    from glob import glob

    flair_paths = pd.read_csv(path.join(bids_dir, 'conversion_info', 'flair_paths.tsv'), sep='\t')

    flair_paths = flair_paths[flair_paths.Path.notnull()]

    flair_paths['BIDS_SubjID'] = ['sub-ADNI' + s.replace('_', '') for s in flair_paths.Subject_ID.to_list()]
    flair_paths['BIDS_Session'] = ['ses-' + s.replace('bl', 'm00').upper() for s in flair_paths.VISCODE.to_list()]

    count = 0
    count_wrong = 0
    count_eq = 0

    for r in flair_paths.iterrows():
        image = r[1]
        image_dir = path.join(bids_dir, image.BIDS_SubjID, image.BIDS_Session, 'FLAIR')
        image_pattern = path.join(image_dir, '%s_%s_*' % (image.BIDS_SubjID, image.BIDS_Session))
        files_list = glob(image_pattern)

        if sum(['Eq' in f for f in files_list]):
            # print("Eq images for subject %s in session %s" % (image.BIDS_SubjID, image.BIDS_Session))
            # print("('%s', '%s')," % (image.BIDS_SubjID[-8:].replace('S', '_S_'),
            #                          image.BIDS_Session[4:].replace('M00', 'bl').lower()))
            count_eq += 1
            continue

        if not files_list:
            # print("No images for subject %s in session %s" % (image.BIDS_SubjID, image.BIDS_Session))
            count += 1

        elif len(files_list) > 2:
            print("Wrong files count for subject %s in session %s" % (image.BIDS_SubjID, image.BIDS_Session))
            print(files_list)
            count_wrong += 1

    print(count)
    print(count_wrong)
    print(count_eq)



# coding: utf8


import abc

__author__ = "Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Jorge Samper-Gonzalez"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


class MLWorkflow:
    __metaclass__ = abc.ABCMeta

    # def __init__(self, ml_input, ml_validation, ml_algorithm, output_dir):
    #     self._ml_input = ml_input
    #     self._ml_validation = ml_validation
    #     self._ml_algorithm = ml_algorithm
    #     self._output_dir = output_dir

    @abc.abstractmethod
    def run(self):
        pass

    def save_image(self):

        import os
        import pandas as pd

        pd.io.parsers.read_csv(os.path.join(self._output_dir, 'results.tsv'), sep='\t')

    @staticmethod
    def metric_distribution(metric, labels, output_path, num_classes=2, metric_label='balanced accuracy'):
        """

        Distribution plots of various metrics such as balanced accuracy!

        metric is expected to be ndarray of size [num_repetitions, num_datasets]

        """
        import numpy as np
        import matplotlib.pyplot as plt
        from matplotlib import cm
        from matplotlib.backends.backend_pdf import PdfPages

        num_repetitions = metric.shape[0]
        num_datasets = metric.shape[1]
        assert len(labels) == num_datasets, "Differing number of features and labels!"
        method_ticks = 1.0 + np.arange(num_datasets)

        fig, ax = plt.subplots(figsize=[9, 9])
        line_coll = ax.violinplot(metric, widths=0.8, bw_method=0.2,
                                  showmedians=True, showextrema=False,
                                  positions=method_ticks)

        cmap = cm.get_cmap('Paired', num_datasets)
        for cc, ln in enumerate(line_coll['bodies']):
            ln.set_facecolor(cmap(cc))
            ln.set_label(labels[cc])

        plt.legend(loc=2, ncol=num_datasets)

        ax.tick_params(axis='both', which='major', labelsize=15)
        ax.grid(axis='y', which='major')

        lower_lim = np.round(np.min([np.float64(0.9 / num_classes), metric.min()]), 3)
        upper_lim = np.round(np.max([1.01, metric.max()]), 3)
        step_tick = 0.1
        ax.set_ylim(lower_lim, upper_lim)

        ax.set_xticks(method_ticks)
        ax.set_xlim(np.min(method_ticks) - 1, np.max(method_ticks) + 1)
        ax.set_xticklabels(labels, rotation=45)  # 'vertical'

        ax.set_yticks(np.arange(lower_lim, upper_lim, step_tick))
        ax.set_yticklabels(np.arange(lower_lim, upper_lim, step_tick))
        # plt.xlabel(xlabel, fontsize=16)
        plt.ylabel(metric_label, fontsize=16)

        fig.tight_layout()

        pp1 = PdfPages(output_path + '.pdf')
        pp1.savefig()
        pp1.close()

        return


class MLInput:
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def get_x(self):
        pass

    @abc.abstractmethod
    def get_y(self):
        pass


class MLValidation:
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def validate(self, y):
        pass


class MLAlgorithm:
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def evaluate(self, train_index, test_index):
        pass

    @abc.abstractmethod
    def save_classifier(self, classifier, output_dir):
        pass

    @abc.abstractmethod
    def save_parameters(self, parameters, output_dir):
        pass



# coding: utf-8

"""
 Module for converting T1 of ADNI
"""

__author__ = "Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Sabrina Fontanella"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


def convert_adni_t1(source_dir, csv_dir, dest_dir, subjs_list=None):
    """
    Convert T1 MR images of ADNI into BIDS format

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    """

    from os import path

    from pandas.io import parsers
    from colorama import Fore

    from clinica.utils.stream import cprint
    from clinica.iotools.converters.adni_to_bids.adni_utils import t1_pet_paths_to_bids

    if subjs_list is None:
        adni_merge_path = path.join(csv_dir, 'ADNIMERGE.csv')
        adni_merge = parsers.read_csv(adni_merge_path, sep=',', low_memory=False)
        subjs_list = list(adni_merge.PTID.unique())

    cprint('Calculating paths of T1 images. Output will be stored in ' + path.join(dest_dir, 'conversion_info') + '.')
    images = compute_t1_paths(source_dir, csv_dir, dest_dir, subjs_list)
    cprint('Paths of T1 images found. Exporting images into BIDS ...')
    t1_pet_paths_to_bids(images, dest_dir, 't1')
    cprint(Fore.GREEN + 'T1 conversion done.' + Fore.RESET)


def compute_t1_paths(source_dir, csv_dir, dest_dir, subjs_list):
    """
    Compute the paths to T1 MR images and store them in a tsv file

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    Returns:
        images: a dataframe with all the paths to the T1 MR images that will be converted into BIDS

    """

    import operator
    from functools import reduce
    from os import path, makedirs

    import pandas as pd

    from clinica.utils.stream import cprint
    from clinica.iotools.converters.adni_to_bids.adni_utils import find_image_path

    t1_col_df = ['Subject_ID', 'VISCODE', 'Visit', 'Sequence', 'Scan_Date',
                 'Study_ID', 'Field_Strength', 'Series_ID', 'Image_ID', 'Original']
    t1_df = pd.DataFrame(columns=t1_col_df)

    # Loading needed .csv files
    adni_merge = pd.read_csv(path.join(csv_dir, 'ADNIMERGE.csv'), sep=',', low_memory=False)
    mprage_meta = pd.read_csv(path.join(csv_dir, 'MPRAGEMETA.csv'), sep=',', low_memory=False)
    mri_quality = pd.read_csv(path.join(csv_dir, 'MRIQUALITY.csv'), sep=',', low_memory=False)
    mayo_mri_qc = pd.read_csv(path.join(csv_dir, 'MAYOADIRL_MRI_IMAGEQC_12_08_15.csv'), sep=',', low_memory=False)
    # Keep only T1 scans
    mayo_mri_qc = mayo_mri_qc[mayo_mri_qc.series_type == 'T1']

    # We will convert the images for each subject in the subject list
    for subj in subjs_list:

        # Filter ADNIMERGE, MPRAGE METADATA and QC for only one subject and sort the rows/visits by examination date
        adnimerge_subj = adni_merge[adni_merge.PTID == subj]
        adnimerge_subj = adnimerge_subj.sort_values('EXAMDATE')

        mprage_meta_subj = mprage_meta[mprage_meta.SubjectID == subj]
        mprage_meta_subj = mprage_meta_subj.sort_values('ScanDate')

        mri_quality_subj = mri_quality[mri_quality.RID == int(subj[-4:])]
        mayo_mri_qc_subj = mayo_mri_qc[mayo_mri_qc.RID == int(subj[-4:])]

        # Obtain corresponding timepoints for the subject visits
        visits = visits_to_timepoints_t1(subj, mprage_meta_subj, adnimerge_subj)

        for visit_info in visits.keys():
            cohort = visit_info[1]
            timepoint = visit_info[0]
            visit_str = visits[visit_info]

            image_dict = None

            if cohort in ('ADNI1', 'ADNIGO', 'ADNI2'):
                image_dict = adni1GO2_image(subj, timepoint, visit_str, mprage_meta_subj, mri_quality_subj,
                                            mayo_mri_qc_subj,
                                            preferred_field_strength=1.5 if cohort == 'ADNI1' else 3.0)
            elif cohort == 'ADNI3':
                image_dict = adni3_image(subj, timepoint, visit_str, mprage_meta_subj, mayo_mri_qc_subj)
            else:
                cprint("Subject %s visit %s belongs to an unknown cohort: %s" % (subj, visit_str, cohort))

            if image_dict is None:
                image_dict = {'Subject_ID': subj,
                              'VISCODE': visit_info[0],
                              'Visit': visits[visit_info],
                              'Sequence': '',
                              'Scan_Date': '',
                              'Study_ID': '',
                              'Series_ID': '',
                              'Image_ID': '',
                              'Field_Strength': '',
                              'Original': True}

            row_to_append = pd.DataFrame(image_dict, index=['i', ])
            # TODO Replace iteratively appending by pandas.concat
            t1_df = t1_df.append(row_to_append, ignore_index=True)

    # Exceptions
    # ==========
    conversion_errors = [  # Eq_1
                         ('031_S_0830', 'm48'),
                         ('100_S_0995', 'm18'),
                         ('031_S_0867', 'm48'),
                         ('100_S_0892', 'm18'),
                         # Empty folders
                         ('029_S_0845', 'm24'),
                         ('094_S_1267', 'm24'),
                         ('029_S_0843', 'm24'),
                         ('027_S_0307', 'm48'),
                         ('057_S_1269', 'm24'),
                         ('036_S_4899', 'm03'),
                         ('033_S_1016', 'm120'),
                         ('130_S_4984', 'm12'),
                         ('027_S_4802', 'm06'),
                         ('131_S_0409', 'bl'),
                         ('082_S_4224', 'm24'),
                         ('006_S_4960', 'bl'),
                         ('006_S_4960', 'm03'),
                         ('006_S_4960', 'm06'),
                         ('006_S_4960', 'm12'),
                         ('006_S_4960', 'm24'),
                         ('006_S_4960', 'm36'),
                         ('006_S_4960', 'm72'),
                         ('022_S_5004', 'bl'),
                         ('022_S_5004', 'm03'),
                         # T1wa
                         ('006_S_4485', 'm84')]

    error_indices = []
    for conv_error in conversion_errors:
        error_indices.append((t1_df.Subject_ID == conv_error[0])
                             & (t1_df.VISCODE == conv_error[1]))

    if error_indices:
        indices_to_remove = t1_df.index[reduce(operator.or_, error_indices, False)]
        t1_df.drop(indices_to_remove, inplace=True)

    # Checking for images paths in filesystem
    images = find_image_path(t1_df, source_dir, 'T1', 'S', 'Series_ID')

    # Store the paths inside a file called conversion_info inside the input directory
    t1_tsv_path = path.join(dest_dir, 'conversion_info')
    if not path.exists(t1_tsv_path):
        makedirs(t1_tsv_path)
    images.to_csv(path.join(t1_tsv_path, 't1_paths.tsv'), sep='\t', index=False)

    return images


def adni1GO2_image(subject_id, timepoint, visit_str, mprage_meta_subj, mri_quality_subj, mayo_mri_qc_subj,
                   preferred_field_strength=3.0):

    from clinica.iotools.converters.adni_to_bids.adni_utils import replace_sequence_chars

    # Get the preferred scan (image series that has been Scaled)
    filtered_mprage = mprage_meta_subj[(mprage_meta_subj['Orig/Proc'] == 'Processed')
                                       & (mprage_meta_subj.Visit == visit_str)
                                       & (mprage_meta_subj.Sequence.map(lambda x: x.endswith('Scaled')))]

    # If no preferred image found, get N3 processed image (N3m)
    if filtered_mprage.shape[0] < 1:
        filtered_mprage = mprage_meta_subj[(mprage_meta_subj['Orig/Proc'] == 'Processed')
                                           & (mprage_meta_subj.Visit == visit_str)
                                           & (mprage_meta_subj.Sequence.map(lambda x: x.endswith('N3m')))]

        # If no N3 processed image found (it means there are no processed images at all), get best original image
        if filtered_mprage.shape[0] < 1:
            return original_image(subject_id, timepoint, visit_str, mprage_meta_subj, mayo_mri_qc_subj,
                                  preferred_field_strength)

    # If there are images with different magnetic field strength, prefer 1.5T images for ADNI1, 3.0T otherwise
    if len(filtered_mprage.MagStrength.unique()) > 1:
        filtered_mprage = filtered_mprage[filtered_mprage.MagStrength == preferred_field_strength]

    # Sort by Series ID in case there are several images, so we keep the one acquired first
    filtered_mprage = filtered_mprage.sort_values('SeriesID')

    scan = filtered_mprage.iloc[0]

    # Check if selected scan passes QC (if QC exists)
    if not check_qc(scan, subject_id, visit_str, mprage_meta_subj, mri_quality_subj):
        return None

    n3 = scan.Sequence.find('N3')
    # Sequence ends in 'N3' or in 'N3m'
    sequence = scan.Sequence[:n3 + 2 + int(scan.Sequence[n3 + 2] == 'm')]
    sequence = replace_sequence_chars(sequence)

    return {'Subject_ID': subject_id,
            'VISCODE': timepoint,
            'Visit': visit_str,
            'Sequence': sequence,
            'Scan_Date': scan.ScanDate,
            'Study_ID': str(scan.StudyID),
            'Series_ID': str(scan.SeriesID),
            'Image_ID': str(scan.ImageUID),
            'Field_Strength': scan.MagStrength,
            'Original': False}


def original_image(subject_id, timepoint, visit_str, mprage_meta_subj, mayo_mri_qc_subj, preferred_field_strength=3.0):

    from clinica.iotools.converters.adni_to_bids.adni_utils import replace_sequence_chars
    from clinica.utils.stream import cprint

    mprage_meta_subj_orig = mprage_meta_subj[mprage_meta_subj['Orig/Proc'] == 'Original']

    cond_mprage = ((mprage_meta_subj_orig.Visit == visit_str)
                   & mprage_meta_subj_orig.Sequence.map(
                       lambda x: ((x.lower().find('mprage') > -1)
                                  | (x.lower().find('mp-rage') > -1)
                                  | (x.lower().find('mp rage') > -1))
                       & (x.find('2') < 0)
                       )
                   )

    cond_spgr = ((mprage_meta_subj_orig.Visit == visit_str)
                 & mprage_meta_subj_orig.Sequence.map(
                    lambda x: (x.lower().find('spgr') > -1)
                    & (x.lower().find('acc') < 0)
                    )
                 )

    filtered_scan = mprage_meta_subj_orig[cond_mprage | cond_spgr]

    if filtered_scan.shape[0] < 1:
        # TODO - LOG THIS
        cprint('NO MPRAGE Meta: ' + subject_id + ' for visit ' + timepoint + ' - ' + visit_str)
        return None

    scan = select_scan_qc_adni2(filtered_scan, mayo_mri_qc_subj, preferred_field_strength)
    sequence = replace_sequence_chars(scan.Sequence)

    return {'Subject_ID': subject_id,
            'VISCODE': timepoint,
            'Visit': visit_str,
            'Sequence': sequence,
            'Scan_Date': scan.ScanDate,
            'Study_ID': str(scan.StudyID),
            'Series_ID': str(scan.SeriesID),
            'Image_ID': str(scan.ImageUID),
            'Field_Strength': scan.MagStrength,
            'Original': True}


def adni3_image(subject_id, timepoint, visit_str, mprage_meta_subj, mayo_mri_qc_subj):

    from clinica.iotools.converters.adni_to_bids.adni_utils import replace_sequence_chars
    from clinica.utils.stream import cprint

    filtered_scan = mprage_meta_subj[(mprage_meta_subj['Orig/Proc'] == 'Original')
                                     & (mprage_meta_subj.Visit == visit_str)
                                     & mprage_meta_subj.Sequence.map(
        lambda x: (x.lower().find('accel') > -1) & ~(x.lower().endswith('_ND')))]

    if filtered_scan.shape[0] < 1:
        # TODO - LOG THIS
        cprint('NO MPRAGE Meta for ADNI3: ' + subject_id + ' for visit ' + timepoint + ' - ' + visit_str)
        return None

    scan = select_scan_qc_adni2(filtered_scan, mayo_mri_qc_subj, preferred_field_strength=3.0)
    sequence = replace_sequence_chars(scan.Sequence)

    return {'Subject_ID': subject_id,
            'VISCODE': timepoint,
            'Visit': visit_str,
            'Sequence': sequence,
            'Scan_Date': scan.ScanDate,
            'Study_ID': str(scan.StudyID),
            'Series_ID': str(scan.SeriesID),
            'Image_ID': str(scan.ImageUID),
            'Field_Strength': scan.MagStrength,
            'Original': True}


def check_qc(scan, subject_id, visit_str, mprage_meta_subj, mri_quality_subj):
    from clinica.utils.stream import cprint

    series_id = scan.SeriesID
    qc_passed = True

    # Check if QC exists for image series
    qc = mri_quality_subj[mri_quality_subj.LONIUID == 'S' + str(scan.SeriesID)]

    # If QC exists and failed we keep the other scan (in case 2 scans were performed)
    if qc.shape[0] > 0 and qc.iloc[0].PASS != 1:
        cprint('QC found but NOT passed')
        cprint('Subject ' + subject_id + ' - Series: ' + str(scan.SeriesID) + ' - Study: ' + str(scan.StudyID))

        mprage_meta_subj_alt = mprage_meta_subj[(mprage_meta_subj['Orig/Proc'] == 'Original')
                                                & (mprage_meta_subj.Visit == visit_str)
                                                & (mprage_meta_subj.SeriesID != series_id)]

        qc_prev_sequence = scan.Sequence
        scan = mprage_meta_subj_alt.iloc[0]
        series_id = scan.SeriesID
        qc_passed = False

    if not qc_passed:
        if scan.Sequence == 'MP-RAGE':
            original_img_seq = 'MPR'
        else:  # 'MP-RAGE REPEAT'
            original_img_seq = 'MPR-R'

        processing_seq = qc_prev_sequence[qc_prev_sequence.find(';'):qc_prev_sequence.find('Scaled') - 2]
        sequence = original_img_seq + processing_seq
        # print sequence

    qc = mri_quality_subj[mri_quality_subj.LONIUID == 'S' + str(scan.SeriesID)]
    if qc.shape[0] > 0 and qc.iloc[0].PASS != 1:
        # TODO - LOG THIS
        cprint('QC found but NOT passed')
        cprint('Subject ' + subject_id + ' - Series: ' + str(scan.SeriesID) + ' - Study: ' + str(scan.StudyID))
        return False

    return True


def visits_to_timepoints_t1(subject, mprage_meta_subj_orig, adnimerge_subj):
    from datetime import datetime
    from clinica.iotools.converters.adni_to_bids.adni_utils import days_between
    from clinica.utils.stream import cprint

    mprage_meta_subj_orig = mprage_meta_subj_orig[mprage_meta_subj_orig['Visit'] != 'ADNI Baseline']

    visits = dict()
    unique_visits = list(mprage_meta_subj_orig.Visit.unique())
    pending_timepoints = []

    # We try to obtain the corresponding image Visit for a given VISCODE
    for adni_row in adnimerge_subj.iterrows():
        visit = adni_row[1]

        if visit.ORIGPROT == 'ADNI3':
            if visit.VISCODE == 'bl':
                preferred_visit_name = 'ADNI Screening'
            else:
                year = str(int(visit.VISCODE[1:]) / 12)
                preferred_visit_name = 'ADNI3 Year ' + year + ' Visit'
        elif visit.ORIGPROT == 'ADNI2':
            if visit.VISCODE == 'bl':
                preferred_visit_name = 'ADNI2 Screening MRI-New Pt'
            elif visit.VISCODE == 'm03':
                preferred_visit_name = 'ADNI2 Month 3 MRI-New Pt'
            elif visit.VISCODE == 'm06':
                preferred_visit_name = 'ADNI2 Month 6-New Pt'
            else:
                year = str(int(visit.VISCODE[1:]) / 12)
                preferred_visit_name = 'ADNI2 Year ' + year + ' Visit'
        else:
            if visit.VISCODE == 'bl':
                if visit.ORIGPROT == 'ADNI1':
                    preferred_visit_name = 'ADNI Screening'
                else:  # ADNIGO
                    preferred_visit_name = 'ADNIGO Screening MRI'
            elif visit.VISCODE == 'm03':  # Only for ADNIGO Month 3
                preferred_visit_name = 'ADNIGO Month 3 MRI'
            else:
                month = int(visit.VISCODE[1:])
                if month < 54:
                    preferred_visit_name = 'ADNI1/GO Month ' + str(month)
                else:
                    preferred_visit_name = 'ADNIGO Month ' + str(month)

        if preferred_visit_name in unique_visits:
            key_preferred_visit = (visit.VISCODE, visit.COLPROT, visit.ORIGPROT)
            if key_preferred_visit not in visits.keys():
                visits[key_preferred_visit] = preferred_visit_name
            elif visits[key_preferred_visit] != preferred_visit_name:
                cprint('[T1] Subject ' + subject + ' has multiple visits for one timepoint. ')
                # cprint(key_preferred_visit)
                # cprint(visits[key_preferred_visit])
                # cprint(visit)
            unique_visits.remove(preferred_visit_name)
            continue

        pending_timepoints.append(visit)

    # Then for images.Visit non matching the expected labels we find the closest date in visits list
    for visit in unique_visits:
        image = (mprage_meta_subj_orig[mprage_meta_subj_orig.Visit == visit]).iloc[0]
        min_db = 100000
        min_db2 = 0
        min_visit = None
        min_visit2 = None

        for timepoint in pending_timepoints:
            db = days_between(image.ScanDate, timepoint.EXAMDATE)
            if db < min_db:
                min_db2 = min_db
                min_visit2 = min_visit

                min_db = db
                min_visit = timepoint

        if min_visit is None:
            cprint('No corresponding timepoint in ADNIMERGE for subject ' + subject + ' in visit ' + image.Visit)
            cprint(image)
            continue

        if min_visit2 is not None and min_db > 90:
            cprint('More than 60 days for corresponding timepoint in ADNIMERGE for subject ' + subject + ' in visit ' + image.Visit + ' on ' + image.ScanDate)
            cprint('Timepoint 1: ' + min_visit.VISCODE + ' - ' + min_visit.ORIGPROT + ' on ' + min_visit.EXAMDATE + ' (Distance: ' + str(
                min_db) + ' days)')
            cprint('Timepoint 2: ' + min_visit2.VISCODE + ' - ' + min_visit2.ORIGPROT + ' on ' + min_visit2.EXAMDATE + ' (Distance: ' + str(
                min_db2) + ' days)')

            # If image is too close to the date between two visits we prefer the earlier visit
            if (datetime.strptime(min_visit.EXAMDATE, "%Y-%m-%d")
                    > datetime.strptime(image.ScanDate, "%Y-%m-%d")
                    > datetime.strptime(min_visit2.EXAMDATE, "%Y-%m-%d")):
                dif = days_between(min_visit.EXAMDATE, min_visit2.EXAMDATE)
                if abs((dif / 2.0) - min_db) < 30:
                    min_visit = min_visit2

            cprint('We prefer ' + min_visit.VISCODE)

        key_min_visit = (min_visit.VISCODE, min_visit.COLPROT, min_visit.ORIGPROT)
        if key_min_visit not in visits.keys():
            visits[key_min_visit] = image.Visit
        elif visits[key_min_visit] != image.Visit:
            cprint('[T1] Subject ' + subject + ' has multiple visits for one timepoint.')
            # cprint(key_min_visit)
            # cprint(visits[key_min_visit])
            # cprint(image.Visit)

    return visits


def select_scan_no_qc(scans_meta):

    selected_scan = scans_meta[scans_meta.Sequence.map(
        lambda x: (x.lower().find('repeat') < 0))]
    if selected_scan.shape[0] < 1:
        selected_scan = scans_meta

    scan = selected_scan.iloc[0]
    return scan


def select_scan_qc_adni2(scans_meta, mayo_mri_qc_subj, preferred_field_strength):
    import numpy as np

    multiple_mag_strength = False
    # Select preferred_field_strength images
    if len(scans_meta.MagStrength.unique()) > 1:
        multiple_mag_strength = True
        not_preferred_scan = scans_meta[scans_meta.MagStrength != preferred_field_strength]
        scans_meta = scans_meta[scans_meta.MagStrength == preferred_field_strength]

    if scans_meta.MagStrength.unique()[0] == 3.0:

        id_list = scans_meta.ImageUID.unique()
        image_ids = ['I' + str(imageuid) for imageuid in id_list]
        int_ids = [int(imageuid) for imageuid in id_list]
        images_qc = mayo_mri_qc_subj[mayo_mri_qc_subj.loni_image.isin(image_ids)]

        if images_qc.shape[0] > 0:
            selected_image = None
            if np.sum(images_qc.series_selected) == 1:
                selected_image = images_qc[images_qc.series_selected == 1].iloc[0].loni_image[1:]
            else:
                images_not_rejected = images_qc[images_qc.series_quality < 4]

                if images_not_rejected.shape[0] < 1:
                    # There are no images that passed the qc
                    # so we'll try to see if there are other images without qc,
                    # otherwise return None
                    qc_ids = set([int(qc_id[1:]) for qc_id in images_qc.loni_image.unique()])
                    no_qc_ids = list(set(int_ids) - qc_ids)
                    # If none of images passed qc the scan is None, otherwise:
                    if len(no_qc_ids) > 0:
                        no_qc_scans_meta = scans_meta[scans_meta.ImageUID.isin(no_qc_ids)]
                        return select_scan_no_qc(no_qc_scans_meta)
                else:
                    series_quality = [q if q > 0 else 4 for q in list(images_not_rejected.series_quality)]
                    best_q = np.amin(series_quality)
                    if best_q == 4:
                        best_q = -1
                    images_best_qc = images_not_rejected[images_not_rejected.series_quality == best_q]
                    if images_best_qc.shape[0] == 1:
                        selected_image = images_best_qc.iloc[0].loni_image[1:]
                    else:
                        best_ids = [int(x[1:]) for x in images_best_qc.loni_image.unique()]
                        best_qc_meta = scans_meta[scans_meta.ImageUID.isin(best_ids)]
                        return select_scan_no_qc(best_qc_meta)

            if selected_image is None and multiple_mag_strength:
                scans_meta = not_preferred_scan
            else:
                scan = scans_meta[scans_meta.ImageUID == int(selected_image)].iloc[0]
                return scan

    # 1.5T or 3.0T without QC
    return select_scan_no_qc(scans_meta)


def check_exceptions(bids_dir):
    from os import path
    import pandas as pd
    from glob import glob

    t1_paths = pd.read_csv(path.join(bids_dir, 'conversion_info', 't1_paths.tsv'), sep='\t')

    t1_paths = t1_paths[t1_paths.Path.notnull()]

    t1_paths['BIDS_SubjID'] = ['sub-ADNI' + s.replace('_', '') for s in t1_paths.Subject_ID.to_list()]
    t1_paths['BIDS_Session'] = ['ses-' + s.replace('bl', 'm00').upper() for s in t1_paths.VISCODE.to_list()]

    count = 0

    for r in t1_paths.iterrows():
        image = r[1]
        image_dir = path.join(bids_dir, image.BIDS_SubjID, image.BIDS_Session, 'anat')
        image_pattern = path.join(image_dir, '*')  # % (image.BIDS_SubjID, image.BIDS_Session))
        files_list = glob(image_pattern)
        if not files_list:
            print("No images for subject %s in session %s" % (image.BIDS_SubjID, image.BIDS_Session))
            count += 1

        elif len(files_list) > 1:
            print("Too many images for subject %s in session %s" % (image.BIDS_SubjID, image.BIDS_Session))
            print(files_list)

    print(count)



"""
Script that generate the skeleton for a new Pipeline
"""
from clinica.engine.cmdparser import CmdParser
from os.path import join, isdir, splitext
from os import mkdir, getcwd


class CmdGenerateTemplates(CmdParser):
    """
    Using the jinja2 library and the pipelines template's file (in the resource dir)
    the user can create his own run-to-go pipelines
    """

    def __init__(self):
        CmdParser.__init__(self)

    def define_name(self):
        self._name = 'template'

    def define_description(self):
        self._description = 'Generate the skeleton for a new pipeline (for developers)'

    def define_options(self):
        self._args.add_argument("name",
                                help='The pipelines title')
        self._args.add_argument("-d", "--output_dir",
                                help='Define the path where generate the directory')

    def run_command(self, args):
        # Parsing input arguments
        if args.output_dir is None:
            args.output_dir = getcwd()
        pipeline = dict()
        pipeline['title'] = args.name
        pipeline['module_name'] = pipeline['title'].replace(' ', '_').lower()
        pipeline['class_name'] = pipeline['title'].replace(' ', '')
        pipeline['command_name'] = pipeline['title'].replace(' ', '-').lower()
        pipeline['dir'] = join(args.output_dir, pipeline['module_name'])

        from jinja2 import Environment, PackageLoader, select_autoescape
        env = Environment(
            loader=PackageLoader('clinica', 'resources/templates/pipeline_template'),
            autoescape=select_autoescape(['py'])
        )

        if not isdir(pipeline['dir']):
            mkdir(pipeline['dir'])

        for template_file in env.list_templates(extensions='j2'):
            template = env.get_template(template_file)
            rendered_file, template_ext = splitext(template_file)
            rendered_filename, rendered_file_ext = splitext(rendered_file)
            if rendered_file_ext == '.py' and rendered_filename != '__init__':
                path_to_write = join(pipeline['dir'], pipeline['module_name'] + '_' + rendered_file)
            else:
                path_to_write = join(pipeline['dir'], rendered_file)
            with open(path_to_write, 'w+') as file_to_write:
                print("Generating template %s" % file_to_write.name)
                file_to_write.write(template.render(pipeline=pipeline))



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
# -*- coding: utf-8 -*-
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

import os.path as op

from nipype.interfaces.base import (CommandLineInputSpec, CommandLine, traits, TraitedSpec,
                                    File, InputMultiPath, isdefined)
from nipype.interfaces.mrtrix3.base import MRTrix3BaseInputSpec, MRTrix3Base


class TensorMetricsInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr='%s',
        mandatory=True,
        position=-1,
        desc='input DTI image')

    out_fa = File(argstr='-fa %s', desc='output FA file')
    out_adc = File(argstr='-adc %s', desc='output ADC file')
    out_ad = File(argstr='-ad %s', desc='output AD file')
    out_rd = File(argstr='-rd %s', desc='output RD file')
    out_evec = File(
        argstr='-vector %s', desc='output selected eigenvector(s) file')
    out_eval = File(
        argstr='-value %s', desc='output selected eigenvalue(s) file')
    component = traits.List(
        [1],
        usedefault=True,
        argstr='-num %s',
        sep=',',
        desc=('specify the desired eigenvalue/eigenvector(s). Note that '
              'several eigenvalues can be specified as a number sequence'))
    in_mask = File(
        exists=True,
        argstr='-mask %s',
        desc=('only perform computation within the specified binary'
              ' brain mask image'))
    modulate = traits.Enum(
        'FA',
        'none',
        'eval',
        argstr='-modulate %s',
        desc=('how to modulate the magnitude of the'
              ' eigenvectors'))


class TensorMetricsOutputSpec(TraitedSpec):
    out_fa = File(desc='output FA file')
    out_adc = File(desc='output ADC file')
    out_ad = File(desc='output AD file')
    out_rd = File(desc='output RD file')
    out_evec = File(desc='output selected eigenvector(s) file')
    out_eval = File(desc='output selected eigenvalue(s) file')


class TensorMetrics(CommandLine):
    """
    Compute metrics from tensors


    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> comp = mrt.TensorMetrics()
    >>> comp.inputs.in_file = 'dti.mif'
    >>> comp.inputs.out_fa = 'fa.mif'
    >>> comp.cmdline                               # doctest: +ELLIPSIS
    'tensor2metric -num 1 -fa fa.mif dti.mif'
    >>> comp.run()                                 # doctest: +SKIP
    """

    _cmd = 'tensor2metric'
    input_spec = TensorMetricsInputSpec
    output_spec = TensorMetricsOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()

        for k in list(outputs.keys()):
            if isdefined(getattr(self.inputs, k)):
                outputs[k] = op.abspath(getattr(self.inputs, k))

        return outputs



# coding: utf8


import os
from os import path
import numpy as np

from clinica.pipelines.machine_learning import base, input, algorithm, validation

__author__ = "Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Jorge Samper-Gonzalez"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"

# This code is an example of implementation of machine learning pipelines


class VB_KFold_DualSVM(base.MLWorkflow):

    # First of all, input has to be chosen. According to it (CAPSVoxelBasedInput or CAPSRegionBasedInput),
    # all the necessary inputs can be found in input.py

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, output_dir, fwhm=0,
                 modulated="on", pvc=None, precomputed_kernel=None, mask_zeros=True, n_threads=15, n_folds=10,
                 grid_search_folds=10, balanced=True, c_range=np.logspace(-6, 2, 17)):

        # Here some parameters selected for this task

        self._output_dir = output_dir
        self._n_threads = n_threads
        self._n_folds = n_folds
        self._grid_search_folds = grid_search_folds
        self._balanced = balanced
        self._c_range = c_range

        # In this case we are running a voxel based input approach
        #

        self._input = input.CAPSVoxelBasedInput(caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id,
                                                image_type, fwhm, modulated, pvc, mask_zeros, precomputed_kernel)

        # Validation and algorithm will be selected in the next part of code

        self._validation = None
        self._algorithm = None

    def run(self):

        # Call on parameters already computed

        x = self._input.get_x()
        y = self._input.get_y()
        kernel = self._input.get_kernel()

        # Now algorithm has been selected, in this case Dual SVM algorithm.
        # Look at algorithm.py to understand the input necessary for each method
        # input parameters were chosen previously

        self._algorithm = algorithm.DualSVMAlgorithm(kernel,
                                                     y,
                                                     balanced=self._balanced,
                                                     grid_search_folds=self._grid_search_folds,
                                                     c_range=self._c_range,
                                                     n_threads=self._n_threads)
        # Here validation type is selected, it's the K fold cross-validation

        self._validation = validation.KFoldCV(self._algorithm)

        classifier, best_params, results = self._validation.validate(y, n_folds=self._n_folds, n_threads=self._n_threads)

        # Creation of the path where all the results will be saved

        classifier_dir = path.join(self._output_dir, 'classifier')
        if not path.exists(classifier_dir):
            os.makedirs(classifier_dir)

        # Here we have selected whant we wanted save
        self._algorithm.save_classifier(classifier, classifier_dir)
        self._algorithm.save_weights(classifier, x, classifier_dir)
        self._algorithm.save_parameters(best_params, classifier_dir)

        self._validation.save_results(self._output_dir)

        # self._input.save_weights_as_nifti(weights)


class VB_RepKFold_DualSVM(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, output_dir, fwhm=0,
                 modulated="on", precomputed_kernel=None, mask_zeros=True, n_threads=15, n_iterations=100, n_folds=10,
                 grid_search_folds=10, balanced=True, c_range=np.logspace(-6, 2, 17)):
        self._output_dir = output_dir
        self._n_threads = n_threads
        self._n_iterations = n_iterations
        self._n_folds = n_folds
        self._grid_search_folds = grid_search_folds
        self._balanced = balanced
        self._c_range = c_range

        self._input = input.CAPSVoxelBasedInput(caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id,
                                                image_type, fwhm, modulated, mask_zeros, precomputed_kernel)
        self._validation = None
        self._algorithm = None

    def run(self):

        x = self._input.get_x()
        y = self._input.get_y()
        kernel = self._input.get_kernel()

        self._algorithm = algorithm.DualSVMAlgorithm(kernel,
                                                     y,
                                                     balanced=self._balanced,
                                                     grid_search_folds=self._grid_search_folds,
                                                     c_range=self._c_range,
                                                     n_threads=self._n_threads)

        self._validation = validation.RepeatedKFoldCV(self._algorithm)

        classifier, best_params, results = self._validation.validate(y, n_iterations=self._n_iterations,
                                                                     n_folds=self._n_folds, n_threads=self._n_threads)

        classifier_dir = path.join(self._output_dir, 'classifier')
        if not path.exists(classifier_dir):
            os.makedirs(classifier_dir)

        self._algorithm.save_classifier(classifier, classifier_dir)
        weights = self._algorithm.save_weights(classifier, x, classifier_dir)
        self._algorithm.save_parameters(best_params, classifier_dir)

        self._validation.save_results(self._output_dir)

        self._input.save_weights_as_nifti(weights, classifier_dir)


class VB_RepHoldOut_DualSVM(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, output_dir, fwhm=0,
                 modulated="on", pvc=None, precomputed_kernel=None, mask_zeros=True, n_threads=15, n_iterations=100,
                 test_size=0.3, grid_search_folds=10, balanced=True, c_range=np.logspace(-6, 2, 17), splits_indices=None):
        self._output_dir = output_dir
        self._n_threads = n_threads
        self._n_iterations = n_iterations
        self._test_size = test_size
        self._grid_search_folds = grid_search_folds
        self._balanced = balanced
        self._c_range = c_range
        self._splits_indices = splits_indices

        self._input = input.CAPSVoxelBasedInput(caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id,
                                                image_type, fwhm, modulated, pvc, mask_zeros, precomputed_kernel)

        self._validation = None
        self._algorithm = None

    def run(self):

        x = self._input.get_x()
        y = self._input.get_y()
        kernel = self._input.get_kernel()

        self._algorithm = algorithm.DualSVMAlgorithm(kernel,
                                                     y,
                                                     balanced=self._balanced,
                                                     grid_search_folds=self._grid_search_folds,
                                                     c_range=self._c_range,
                                                     n_threads=self._n_threads)

        self._validation = validation.RepeatedHoldOut(self._algorithm, n_iterations=self._n_iterations, test_size=self._test_size)

        classifier, best_params, results = self._validation.validate(y, n_threads=self._n_threads, splits_indices=self._splits_indices)
        classifier_dir = path.join(self._output_dir, 'classifier')
        if not path.exists(classifier_dir):
            os.makedirs(classifier_dir)

        self._algorithm.save_classifier(classifier, classifier_dir)
        self._algorithm.save_parameters(best_params, classifier_dir)
        weights = self._algorithm.save_weights(classifier, x, classifier_dir)

        self._input.save_weights_as_nifti(weights, classifier_dir)

        self._validation.save_results(self._output_dir)


class VertexB_RepHoldOut_dualSVM(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, output_dir, image_type='fdg', fwhm=20,
                 precomputed_kernel=None, n_threads=15, n_iterations=100, test_size=0.3, grid_search_folds=10,
                 balanced=True, c_range=np.logspace(-10, 2, 1000), splits_indices=None):

        self._output_dir = output_dir
        self._n_threads = n_threads
        self._n_iterations = n_iterations
        self._test_size = test_size
        self._grid_search_folds = grid_search_folds
        self._balanced = balanced
        self._c_range = c_range
        self._splits_indices = splits_indices

        self._input = input.CAPSVertexBasedInput(caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, fwhm,
                                                 image_type, precomputed_kernel)

        self._validation = None
        self._algorithm = None

    def run(self):

        x = self._input.get_x()
        y = self._input.get_y()
        kernel = self._input.get_kernel()

        self._algorithm = algorithm.DualSVMAlgorithm(kernel,
                                                     y,
                                                     balanced=self._balanced,
                                                     grid_search_folds=self._grid_search_folds,
                                                     c_range=self._c_range,
                                                     n_threads=self._n_threads)

        self._validation = validation.RepeatedHoldOut(self._algorithm, n_iterations=self._n_iterations, test_size=self._test_size)
        classifier, best_params, results = self._validation.validate(y, n_threads=self._n_threads, splits_indices=self._splits_indices)
        classifier_dir = path.join(self._output_dir, 'classifier')
        if not path.exists(classifier_dir):
            os.makedirs(classifier_dir)
        self._algorithm.save_classifier(classifier, classifier_dir)
        self._algorithm.save_parameters(best_params, classifier_dir)
        weights = self._algorithm.save_weights(classifier, x, classifier_dir)
        self._input.save_weights_as_datasurface(weights, classifier_dir)
        self._validation.save_results(self._output_dir)


class RB_RepHoldOut_DualSVM(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type,  atlas,
                 output_dir, pvc=None, n_threads=15, n_iterations=100, test_size=0.3,
                 grid_search_folds=10, balanced=True, c_range=np.logspace(-6, 2, 17), splits_indices=None):
        self._output_dir = output_dir
        self._n_threads = n_threads
        self._n_iterations = n_iterations
        self._test_size = test_size
        self._grid_search_folds = grid_search_folds
        self._balanced = balanced
        self._c_range = c_range
        self._splits_indices = splits_indices

        self._input = input.CAPSRegionBasedInput(caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id,
                                                 image_type, atlas, pvc)
        self._validation = None
        self._algorithm = None

    def run(self):

        x = self._input.get_x()
        y = self._input.get_y()
        kernel = self._input.get_kernel()

        self._algorithm = algorithm.DualSVMAlgorithm(kernel,
                                                     y,
                                                     balanced=self._balanced,
                                                     grid_search_folds=self._grid_search_folds,
                                                     c_range=self._c_range,
                                                     n_threads=self._n_threads)

        self._validation = validation.RepeatedHoldOut(self._algorithm, n_iterations=self._n_iterations, test_size=self._test_size)

        classifier, best_params, results = self._validation.validate(y, n_threads=self._n_threads, splits_indices=self._splits_indices)
        classifier_dir = path.join(self._output_dir, 'classifier')
        if not path.exists(classifier_dir):
            os.makedirs(classifier_dir)

        self._algorithm.save_classifier(classifier, classifier_dir)
        self._algorithm.save_parameters(best_params, classifier_dir)
        weights = self._algorithm.save_weights(classifier, x, classifier_dir)

        self._input.save_weights_as_nifti(weights, classifier_dir)

        self._validation.save_results(self._output_dir)


class RB_RepHoldOut_LogisticRegression(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, atlas,
                 output_dir, pvc=None, n_threads=15, n_iterations=100, test_size=0.3,
                 grid_search_folds=10, balanced=True, c_range=np.logspace(-6, 2, 17), splits_indices=None):
        self._output_dir = output_dir
        self._n_threads = n_threads
        self._n_iterations = n_iterations
        self._test_size = test_size
        self._grid_search_folds = grid_search_folds
        self._balanced = balanced
        self._c_range = c_range
        self._splits_indices = splits_indices

        self._input = input.CAPSRegionBasedInput(caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id,
                                                 image_type, atlas, pvc)
        self._validation = None
        self._algorithm = None

    def run(self):

        x = self._input.get_x()
        y = self._input.get_y()

        self._algorithm = algorithm.LogisticReg(x, y, balanced=self._balanced,
                                                grid_search_folds=self._grid_search_folds,
                                                c_range=self._c_range,
                                                n_threads=self._n_threads)

        self._validation = validation.RepeatedHoldOut(self._algorithm, n_iterations=self._n_iterations, test_size=self._test_size)
        classifier, best_params, results = self._validation.validate(y, n_threads=self._n_threads, splits_indices=self._splits_indices)

        classifier_dir = os.path.join(self._output_dir, 'classifier')
        if not path.exists(classifier_dir):
            os.makedirs(classifier_dir)

        self._algorithm.save_classifier(classifier, classifier_dir)
        self._algorithm.save_parameters(best_params, classifier_dir)
        weights = self._algorithm.save_weights(classifier, classifier_dir)

        self._classifier = classifier

        self._input.save_weights_as_nifti(weights, classifier_dir)

        self._validation.save_results(self._output_dir)


class RB_RepHoldOut_RandomForest(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, atlas,
                 output_dir, pvc=None, n_threads=15, n_iterations=100, test_size=0.3,
                 grid_search_folds=10, balanced=True, n_estimators_range=(100, 200, 400),
                 max_depth_range=[None], min_samples_split_range=[2],
                 max_features_range=('auto', 0.25, 0.5), splits_indices=None):

        self._output_dir = output_dir
        self._n_threads = n_threads
        self._n_iterations = n_iterations
        self._test_size = test_size
        self._grid_search_folds = grid_search_folds
        self._balanced = balanced
        self._n_estimators_range = n_estimators_range
        self._max_depth_range = max_depth_range
        self._min_samples_split_range = min_samples_split_range
        self._max_features_range = max_features_range
        self._splits_indices = splits_indices

        self._input = input.CAPSRegionBasedInput(caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id,
                                                 image_type, atlas, pvc)
        self._validation = None
        self._algorithm = None

    def run(self):

        x = self._input.get_x()
        y = self._input.get_y()

        self._algorithm = algorithm.RandomForest(x, y, balanced=self._balanced,
                                                 grid_search_folds=self._grid_search_folds,
                                                 n_estimators_range=self._n_estimators_range,
                                                 max_depth_range=self._max_depth_range,
                                                 min_samples_split_range=self._min_samples_split_range,
                                                 max_features_range=self._max_features_range,
                                                 n_threads=self._n_threads)

        self._validation = validation.RepeatedHoldOut(self._algorithm, n_iterations=self._n_iterations, test_size=self._test_size)
        classifier, best_params, results = self._validation.validate(y, n_threads=self._n_threads, splits_indices=self._splits_indices)

        classifier_dir = os.path.join(self._output_dir, 'classifier')
        if not path.exists(classifier_dir):
            os.makedirs(classifier_dir)

        self._algorithm.save_classifier(classifier, classifier_dir)
        self._algorithm.save_parameters(best_params, classifier_dir)
        weights = self._algorithm.save_weights(classifier, classifier_dir)

        self._input.save_weights_as_nifti(weights, classifier_dir)

        self._validation.save_results(self._output_dir)


class RB_LearningCurveRepHoldOut_DualSVM(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type,  atlas,
                 output_dir, pvc=None, precomputed_kernel=None, n_threads=15, n_iterations=100, test_size=0.3,
                 n_learning_points=10, grid_search_folds=10, balanced=True, c_range=np.logspace(-6, 2, 17)):

        self._output_dir = output_dir
        self._n_threads = n_threads
        self._n_iterations = n_iterations
        self._test_size = test_size
        self._n_learning_points = n_learning_points
        self._grid_search_folds = grid_search_folds
        self._balanced = balanced
        self._c_range = c_range

        self._input = input.CAPSRegionBasedInput(caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id,
                                                 image_type, atlas, pvc, precomputed_kernel)
        self._validation = None
        self._algorithm = None

    def run(self):

        x = self._input.get_x()
        y = self._input.get_y()
        kernel = self._input.get_kernel()

        self._algorithm = algorithm.DualSVMAlgorithm(kernel,
                                                     y,
                                                     balanced=self._balanced,
                                                     grid_search_folds=self._grid_search_folds,
                                                     c_range=self._c_range,
                                                     n_threads=self._n_threads)

        self._validation = validation.LearningCurveRepeatedHoldOut(self._algorithm,
                                                                   n_iterations=self._n_iterations,
                                                                   test_size=self._test_size,
                                                                   n_learning_points=self._n_learning_points)

        classifier, best_params, results = self._validation.validate(y, n_threads=self._n_threads)

        for learning_point in range(self._n_learning_points):

            learning_point_dir = path.join(self._output_dir, 'learning_split-' + str(learning_point))

            classifier_dir = path.join(learning_point_dir, 'classifier')
            if not path.exists(classifier_dir):
                os.makedirs(classifier_dir)

            self._algorithm.save_classifier(classifier[learning_point], classifier_dir)
            self._algorithm.save_parameters(best_params[learning_point], classifier_dir)
            weights = self._algorithm.save_weights(classifier[learning_point], x, classifier_dir)

            self._input.save_weights_as_nifti(weights, classifier_dir)

        self._validation.save_results(self._output_dir)


class VB_LearningCurveRepHoldOut_DualSVM(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, output_dir, fwhm=0,
                 modulated="on", pvc=None, precomputed_kernel=None, mask_zeros=True, n_threads=15, n_iterations=100,
                 test_size=0.3, n_learning_points=10, grid_search_folds=10, balanced=True, c_range=np.logspace(-6, 2, 17)):
        self._output_dir = output_dir
        self._n_threads = n_threads
        self._n_iterations = n_iterations
        self._test_size = test_size
        self._n_learning_points = n_learning_points
        self._grid_search_folds = grid_search_folds
        self._balanced = balanced
        self._c_range = c_range

        self._input = input.CAPSVoxelBasedInput(caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id,
                                                image_type, fwhm, modulated, pvc, mask_zeros, precomputed_kernel)

        self._validation = None
        self._algorithm = None

    def run(self):

        x = self._input.get_x()
        y = self._input.get_y()
        kernel = self._input.get_kernel()

        self._algorithm = algorithm.DualSVMAlgorithm(kernel,
                                                     y,
                                                     balanced=self._balanced,
                                                     grid_search_folds=self._grid_search_folds,
                                                     c_range=self._c_range,
                                                     n_threads=self._n_threads)

        self._validation = validation.LearningCurveRepeatedHoldOut(self._algorithm,
                                                                   n_iterations=self._n_iterations,
                                                                   test_size=self._test_size,
                                                                   n_learning_points=self._n_learning_points)

        classifier, best_params, results = self._validation.validate(y, n_threads=self._n_threads)

        for learning_point in range(self._n_learning_points):

            learning_point_dir = path.join(self._output_dir, 'learning_split-' + str(learning_point))

            classifier_dir = path.join(learning_point_dir, 'classifier')
            if not path.exists(classifier_dir):
                os.makedirs(classifier_dir)

            self._algorithm.save_classifier(classifier[learning_point], classifier_dir)
            self._algorithm.save_parameters(best_params[learning_point], classifier_dir)
            weights = self._algorithm.save_weights(classifier[learning_point], x, classifier_dir)

            self._input.save_weights_as_nifti(weights, classifier_dir)

        self._validation.save_results(self._output_dir)


class RB_RepKFold_DualSVM(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type,  atlas,
                 output_dir, pvc=None, n_threads=15, n_iterations=100, test_size=0.3, n_folds=10,
                 grid_search_folds=10, balanced=True, c_range=np.logspace(-6, 2, 17), splits_indices=None):
        self._output_dir = output_dir
        self._n_threads = n_threads
        self._n_iterations = n_iterations
        self._test_size = test_size
        self._grid_search_folds = grid_search_folds
        self._balanced = balanced
        self._c_range = c_range
        self._n_folds = n_folds
        self._splits_indices = splits_indices

        self._input = input.CAPSRegionBasedInput(caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id,
                                                 image_type, atlas, pvc)
        self._validation = None
        self._algorithm = None

    def run(self):

        x = self._input.get_x()
        y = self._input.get_y()
        kernel = self._input.get_kernel()

        self._algorithm = algorithm.DualSVMAlgorithm(kernel,
                                                     y,
                                                     balanced=self._balanced,
                                                     grid_search_folds=self._grid_search_folds,
                                                     c_range=self._c_range,
                                                     n_threads=self._n_threads)

        self._validation = validation.RepeatedKFoldCV(self._algorithm)

        classifier, best_params, results = self._validation.validate(y, n_iterations=self._n_iterations,
                                                                     n_folds=self._n_folds, n_threads=self._n_threads)

        classifier_dir = path.join(self._output_dir, 'classifier')
        if not path.exists(classifier_dir):
            os.makedirs(classifier_dir)

        self._algorithm.save_classifier(classifier, classifier_dir)
        weights = self._algorithm.save_weights(classifier, x, classifier_dir)
        self._algorithm.save_parameters(best_params, classifier_dir)

        self._validation.save_results(self._output_dir)

        self._input.save_weights_as_nifti(weights, classifier_dir)


class TB_RepHoldOut_DualSVM(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type,  atlas, dataset,
                 output_dir, pvc=None, n_threads=15, n_iterations=100, test_size=0.3,
                 grid_search_folds=10, balanced=True, c_range=np.logspace(-6, 2, 17), splits_indices=None):
        self._output_dir = output_dir
        self._n_threads = n_threads
        self._n_iterations = n_iterations
        self._test_size = test_size
        self._grid_search_folds = grid_search_folds
        self._balanced = balanced
        self._c_range = c_range
        self._splits_indices = splits_indices

        self._input = input.CAPSTSVBasedInput(caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type,
                                              atlas, dataset, pvc)

        self._validation = None
        self._algorithm = None

    def run(self):

        x = self._input.get_x()
        y = self._input.get_y()
        kernel = self._input.get_kernel()

        self._algorithm = algorithm.DualSVMAlgorithm(kernel,
                                                     y,
                                                     balanced=self._balanced,
                                                     grid_search_folds=self._grid_search_folds,
                                                     c_range=self._c_range,
                                                     n_threads=self._n_threads)

        self._validation = validation.RepeatedHoldOut(self._algorithm, n_iterations=self._n_iterations, test_size=self._test_size)

        classifier, best_params, results = self._validation.validate(y, n_threads=self._n_threads, splits_indices=self._splits_indices)
        classifier_dir = path.join(self._output_dir, 'classifier')
        if not path.exists(classifier_dir):
            os.makedirs(classifier_dir)

        self._algorithm.save_classifier(classifier, classifier_dir)
        self._algorithm.save_parameters(best_params, classifier_dir)
        weights = self._algorithm.save_weights(classifier, x, classifier_dir)

        self._validation.save_results(self._output_dir)


class TB_RepHoldOut_RandomForest(base.MLWorkflow):
    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, atlas, dataset,
                 output_dir, pvc=None, n_threads=15, n_iterations=100, test_size=0.3,
                 grid_search_folds=10, balanced=True, n_estimators_range=(100, 200, 400),
                 max_depth_range=[None], min_samples_split_range=[2],
                 max_features_range=('auto', 0.25, 0.5), splits_indices=None):

        self._output_dir = output_dir
        self._n_threads = n_threads
        self._n_iterations = n_iterations
        self._test_size = test_size
        self._grid_search_folds = grid_search_folds
        self._balanced = balanced
        self._n_estimators_range = n_estimators_range
        self._max_depth_range = max_depth_range
        self._min_samples_split_range = min_samples_split_range
        self._max_features_range = max_features_range
        self._splits_indices = splits_indices

        self._input = input.CAPSTSVBasedInput(caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id,
                                              image_type, atlas, dataset, pvc)
        self._validation = None
        self._algorithm = None

    def run(self):

        x = self._input.get_x()
        y = self._input.get_y()

        self._algorithm = algorithm.RandomForest(x, y, balanced=self._balanced,
                                                 grid_search_folds=self._grid_search_folds,
                                                 n_estimators_range=self._n_estimators_range,
                                                 max_depth_range=self._max_depth_range,
                                                 min_samples_split_range=self._min_samples_split_range,
                                                 max_features_range=self._max_features_range,
                                                 n_threads=self._n_threads)

        self._validation = validation.RepeatedHoldOut(self._algorithm, n_iterations=self._n_iterations, test_size=self._test_size)
        classifier, best_params, results = self._validation.validate(y, n_threads=self._n_threads, splits_indices=self._splits_indices)

        classifier_dir = os.path.join(self._output_dir, 'classifier')
        if not path.exists(classifier_dir):
            os.makedirs(classifier_dir)

        self._algorithm.save_classifier(classifier, classifier_dir)
        self._algorithm.save_parameters(best_params, classifier_dir)
        weights = self._algorithm.save_weights(classifier, classifier_dir)

        # self._input.save_weights_as_nifti(weights, classifier_dir)

        # self._validation.save_results(self._output_dir)


# SVM reg

class VBREG_RepKFold_DualSVM(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, output_dir, fwhm=0,
                 modulated="on", pvc=None, precomputed_kernel=None, mask_zeros=True, n_threads=15, n_iterations=100,
                 n_folds=10,
                 test_size=0.1, grid_search_folds=10, balanced=True, c_range=np.logspace(-6, 2, 17),
                 splits_indices=None):

        self._output_dir = output_dir
        self._n_threads = n_threads
        self._n_iterations = n_iterations
        self._test_size = test_size
        self._grid_search_folds = grid_search_folds
        self._balanced = balanced
        self._c_range = c_range
        self._splits_indices = splits_indices
        self._n_folds = n_folds
        self._input = input.CAPSVoxelBasedInputREGSVM(caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id,
                                                      image_type, fwhm, modulated, pvc, mask_zeros, precomputed_kernel)
        self._validation = None
        self._algorithm = None

    def run(self):

        x = self._input.get_x()
        y = self._input.get_y()
        kernel = self._input.get_kernel()

        self._algorithm = algorithm.DualSVMAlgorithm(kernel,
                                                     y,
                                                     balanced=self._balanced,
                                                     grid_search_folds=self._grid_search_folds,
                                                     c_range=self._c_range,
                                                     n_threads=self._n_threads)

        self._validation = validation.RepeatedKFoldCV(self._algorithm)
        print('K fold')

        classifier, best_params, results = self._validation.validate(y, n_iterations=self._n_iterations,
                                                                     n_folds=self._n_folds, n_threads=self._n_threads)

        classifier_dir = path.join(self._output_dir, 'classifier')
        if not path.exists(classifier_dir):
            os.makedirs(classifier_dir)

        self._algorithm.save_classifier(classifier, classifier_dir)
        weights = self._algorithm.save_weights(classifier, x, classifier_dir)
        self._algorithm.save_parameters(best_params, classifier_dir)

        self._validation.save_results(self._output_dir)

        self._input.save_weights_as_nifti(weights, classifier_dir)


class RB_RepHoldOut_RandomForest_Multiclass(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, atlas,
                 output_dir, pvc=None, n_threads=15, n_iterations=100, test_size=0.3,
                 grid_search_folds=10, balanced=True, n_estimators_range=(100, 200, 400),
                 max_depth_range=[None], min_samples_split_range=[2],
                 max_features_range=('auto', 0.25, 0.5), splits_indices=None):

        self._output_dir = output_dir
        self._n_threads = n_threads
        self._n_iterations = n_iterations
        self._test_size = test_size
        self._grid_search_folds = grid_search_folds
        self._balanced = balanced
        self._n_estimators_range = n_estimators_range
        self._max_depth_range = max_depth_range
        self._min_samples_split_range = min_samples_split_range
        self._max_features_range = max_features_range
        self._splits_indices = splits_indices

        self._input = input.CAPSRegionBasedInput(caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id,
                                                 image_type, atlas, pvc)
        self._validation = None
        self._algorithm = None

    def run(self):

        x = self._input.get_x()
        y = self._input.get_y()

        self._algorithm = algorithm.RandomForest(x, y, balanced=self._balanced,
                                                 grid_search_folds=self._grid_search_folds,
                                                 n_estimators_range=self._n_estimators_range,
                                                 max_depth_range=self._max_depth_range,
                                                 min_samples_split_range=self._min_samples_split_range,
                                                 max_features_range=self._max_features_range,
                                                 n_threads=self._n_threads)

        self._validation = validation.RepeatedHoldOut(self._algorithm, n_iterations=self._n_iterations, test_size=self._test_size)
        classifier, best_params, results = self._validation.validate(y, n_threads=self._n_threads, splits_indices=self._splits_indices)

        classifier_dir = os.path.join(self._output_dir, 'classifier')
        if not path.exists(classifier_dir):
            os.makedirs(classifier_dir)

        self._algorithm.save_classifier(classifier, classifier_dir)
        self._algorithm.save_parameters(best_params, classifier_dir)
        weights = self._algorithm.save_weights(classifier, classifier_dir)

        self._input.save_weights_as_nifti(weights, classifier_dir)

        self._validation.save_results(self._output_dir)


class VBREG_RepKfold_SVMOV0(base.MLWorkflow):
    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, output_dir, fwhm=0,
                 modulated="on", pvc=None, precomputed_kernel=None, mask_zeros=True, n_threads=15, n_iterations=100,
                 n_folds=10,
                 test_size=0.3, grid_search_folds=10, balanced=True, c_range=np.logspace(-6, 2, 17),
                 splits_indices=None):

        self._output_dir = output_dir
        self._n_threads = n_threads
        self._n_iterations = n_iterations
        self._test_size = test_size
        self._grid_search_folds = grid_search_folds
        self._balanced = balanced
        self._c_range = c_range
        self._splits_indices = splits_indices
        self._n_folds = n_folds
        self._input = input.CAPSVoxelBasedInputREGSVM(caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id,
                                                      image_type, fwhm, modulated, pvc, mask_zeros, precomputed_kernel)
        self._validation = None
        self._algorithm = None

    def run(self):

        x = self._input.get_x()
        y = self._input.get_y()
        kernel = self._input.get_kernel()

        self._algorithm = algorithm.OneVsOneSVM(kernel,
                                                y,
                                                balanced=self._balanced,
                                                grid_search_folds=self._grid_search_folds,
                                                c_range=self._c_range,
                                                n_threads=self._n_threads)

        self._validation = validation.RepeatedKFoldCV_Multiclass(self._algorithm)

        classifier, best_params, results = self._validation.validate(y, n_iterations=self._n_iterations,
                                                                     n_folds=self._n_folds, n_threads=self._n_threads)

        classifier_dir = path.join(self._output_dir, 'classifier')
        if not path.exists(classifier_dir):
            os.makedirs(classifier_dir)

        self._algorithm.save_parameters(best_params, classifier_dir)

        self._validation.save_results(self._output_dir)


class VBREG_RepKfold_SVMOVR(base.MLWorkflow):
    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, output_dir, fwhm=0,
                 modulated="on", pvc=None, precomputed_kernel=None, mask_zeros=True, n_threads=15, n_iterations=100,
                 n_folds=10,
                 test_size=0.3, grid_search_folds=10, balanced=True, c_range=np.logspace(-6, 2, 17),
                 splits_indices=None):

        self._output_dir = output_dir
        self._n_threads = n_threads
        self._n_iterations = n_iterations
        self._test_size = test_size
        self._grid_search_folds = grid_search_folds
        self._balanced = balanced
        self._c_range = c_range
        self._splits_indices = splits_indices
        self._n_folds = n_folds
        self._input = input.CAPSVoxelBasedInputREGSVM(caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id,
                                                      image_type, fwhm, modulated, pvc, mask_zeros, precomputed_kernel)
        self._validation = None
        self._algorithm = None

    def run(self):

        x = self._input.get_x()
        y = self._input.get_y()
        kernel = self._input.get_kernel()

        self._algorithm = algorithm.OneVsRestSVM(kernel,
                                                 y,
                                                 balanced=self._balanced,
                                                 grid_search_folds=self._grid_search_folds,
                                                 c_range=self._c_range,
                                                 n_threads=self._n_threads)

        self._validation = validation.RepeatedKFoldCV_Multiclass(self._algorithm)

        classifier, best_params, results = self._validation.validate(y, n_iterations=self._n_iterations,
                                                                     n_folds=self._n_folds, n_threads=self._n_threads)

        classifier_dir = path.join(self._output_dir, 'classifier')
        if not path.exists(classifier_dir):
            os.makedirs(classifier_dir)

        self._algorithm.save_parameters(best_params, classifier_dir)

        self._validation.save_results(self._output_dir)



# coding: utf8

import clinica.engine as ce


class NifdToBidsCLI(ce.CmdParser):
    """
    todo:add description
    """

    def define_name(self):
        """Define the sub-command name to run this pipelines.
        """
        self._name = 'nifd-to-bids'

    def define_description(self):
        """Define a description of this pipeline.
        """
        self._description = 'Convert NIFD (http://4rtni-ftldni.ini.usc.edu/) into BIDS.'

    def define_options(self):
        """Define the sub-command arguments
        """
        self._args.add_argument("dataset_directory",
                                help='Path to the NIFD images directory.')
        self._args.add_argument("clinical_data_directory",
                                help='Path to the directory containing the NIFD clinical files. '
                                     '(NIFD_Clinical_Data_2017_final_updated.xlsx, '
                                     'DataDictionary_NIFD_2017.10.18.xlsx, idaSearch_1_17_2019_NIFD_all.csv)')
        # self._args.add_argument("ida_file",
        #                         help='Path to the NIFD ida.tsv file, path/to/file/ida.tsv')
        self._args.add_argument("bids_directory",
                                help='Path to the BIDS directory.')

    def run_command(self, args):

        from clinica.iotools.converters.nifd_to_bids.nifd_to_bids import convert_images, convert_clinical_data
        from clinica.iotools.converter_utils import check_bin
        from clinica.utils.stream import cprint
        from colorama import Fore
        import sys

        missing_bin = check_bin('dcm2niix')
        if missing_bin:
            cprint(Fore.RED + 'dcm2niix is required.'
                   + ' Install it and re-run the converter. You can use the following link :'
                   + Fore.BLUE + '\nhttps://github.com/rordenlab/dcm2niix'
                   + Fore.RESET)
            cprint('Exiting clinica...')
            sys.exit()

        # to_convert = convert_images(args.dataset_directory, args.ida_file, args.bids_directory)
        # convert_clinical_data(args.bids_directory, args.ida_file, args.clinical_data_file, to_convert)
        to_convert = convert_images(args.dataset_directory, args.bids_directory, args.clinical_data_directory)
        convert_clinical_data(args.bids_directory, args.clinical_data_directory, to_convert)
        cprint(Fore.GREEN + 'Conversion to BIDS succeeded' + Fore.RESET)



__author__ = "Adam Wild"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Adam Wild"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Adam Wild"
__email__ = "adam.wild@icm-institute.org"
__status__ = "Development"

from clinica.iotools.converters.nifd_to_bids.utils.conv_image_folders import get_all_med_name, dict_conversion


def break_path(path):
    return path.split('/')


def filter(l_paths, source_dir, descriptors):
    """
    From a list of paths, removes paths that do not lead to an image described by any of the descriptor in descriptors

    Args:
        l_paths: list of paths
        source_dir: path to the NIFD dataset
        descriptors: list of descriptor instances

    Returns:
        list of paths
    """
    medical_images = get_all_med_name(source_dir)

    equivalences = dict_conversion(medical_images, descriptors)
    sol = []
    for path in l_paths:
        for elt in break_path(path):
            if elt in equivalences:
                sol.append(path)
                break
    return sol


def csv_to_path(csv_line, source_dir):
    import os

    def change_format_date(date):
        sol = ""
        new_date = date.split('/')
        if len(new_date[0]) != 2:
            new_date[0] = '0' + new_date[0]
        if len(new_date[1]) != 2:
            new_date[1] = '0' + new_date[1]
        sol = new_date[2] + '-' + new_date[0] + '-' + new_date[1]
        return sol

    split = csv_line.split('\t')
    if split[10] != 'Original' and split[9] == 'MT1; GradWarp; N3m <- Sag 3D MP-RAGE':
        split[9] = 'MT1__GradWarp__N3m'
    split[9] = split[9].replace(' ', '_')
    split[9] = split[9].replace(':', '_')
    split[9] = split[9].replace('(', '_')
    split[9] = split[9].replace(')', '_')
    split[9] = split[9].replace('/', '_')
    split[9] = split[9].replace('MT1;_GradWarp;_N3m_<-_Sag_3D_MP-RAGE', 'MT1__GradWarp__N3m')
    folder = os.path.join(source_dir, split[0], split[9])

    try:
        subfold = [f.path.split('/')[-1] for f in os.scandir(folder) if f.is_dir()]
    except Exception:
        return None
    date = change_format_date(split[5])

    if split[0] == '1_S_0005' and date == '2011-06-13' and split[9] in (
            'NIFD_DTI_b=1000_2.7mm3', 'dti_b=2000_64dir', 'dti_b=0_scans'):
        date = '2011-06-15'
    if split[0] == '1_S_0084' and date == '2012-02-13' and split[9] in (
            'NIFD_DTI_b1000_2.7mm3', 'dti_WIP_b2000_64dir', 'DTI_64_2.2iso_full_ky_fov220',
            'DTI_b=0_2.2iso_full_ky_-_10_acqs',
            'dti_WIP_b0_scans_aah'):
        date = '2012-02-15'
    if split[0] == '1_S_0316' and date == '2013-10-30' and split[9] in (
            'NIFD_DTI_b1000_2.7mm3_511E', 'ep2d-advdiff-511E_b2000_64dir', 'ep2d-advdiff-511E_b0_scan'):
        date = '2013-10-31'

    val = [fold for fold in subfold if fold.startswith(date)]

    if val == []:
        return None
    path_date = os.path.join(folder, val[0])
    full_path = os.path.join(path_date, [f.path for f in os.scandir(path_date) if f.is_dir()][0])
    return full_path


def get_patients_source_files(source_dir, path_ida):
    '''
    Returns a dictionary containing the paths of all Dicom files for a patient

    Args:
        source_dir: path to the NIFD dataset
        path_ida : path to the ida.tsv file

    Returns:
        patients_source_files[subject_ID] = [paths_to_all_medical_images_of_subject]
    '''

    sol = dict()
    fich = open(path_ida, 'r')
    fich.readline()
    for line in fich.readlines():
        patient_id = line.split('\t')[0]
        if patient_id not in sol:
            sol[patient_id] = []
        path = csv_to_path(line, source_dir)
        if path is not None:
            sol[patient_id].append(path)
    return sol


def filter_patients_source_files(patients_files, source_dir, descriptors):
    '''
    Iterates over a dictionary containing all source files for a patient,
    returns the same dictionary with only path to files that could be converted

    Args:
        patients_files: patients_source_files[subject_ID] = [paths_to_all_medical_images_of_subject]
        source_dir: path to the NIFD dataset
        descriptors: list of descriptor instances

    Returns:
        patients_source_files[subject_ID] = [paths_to_medical_images_described_by_at_least_one_descriptor]
    '''
    for patient in patients_files:
        patients_files[patient] = filter(patients_files[patient], source_dir, descriptors)

    return patients_files


def create_folder(path):
    """
    If the provided path does not exist, the required folders are created
    """
    import os
    try:
        os.makedirs(path, exist_ok=True)
    except Exception:
        pass


def extract_date(path):
    """
    Extracts the name of the date associated with the given path

    Args:
        path: Path to a Dicom file

    Returns:
        Date
    """
    path = break_path(path)
    for i in path:
        if i.startswith('20'):
            sol = i.split('_')[0]
            return sol


def extract_name_med_img(path, equivalences):
    """
    Extracts the name of the medical image associated with the given path

    Args:
        path: Path to a Dicom file
        equivalences: dictionary, equivalences['medical_image_name'] = (Descriptor_instance, modalityLabel)

    Returns:
        Medical name
    """
    path = break_path(path)
    for i in path:
        if i in equivalences:
            return i


def print_patients_source_files(dict):
    """
    Pretty printer for a nested dictionary such as patients_source_files

    Args:
        dict: Data structure of the form dict['subject_ID'] = [paths/to/dcm]
    """
    from clinica.utils.stream import cprint
    for key in dict:
        cprint(key)
        for elt in dict[key]:
            cprint('\t' + str(elt))


def print_orderedBIDS(dict):
    """
    Pretty printer for a nested dictionary such as ordered_bids, final_bids

    Args:
        dict: Data structure of the form dict['session_id']['dataType']['Priority']['Final_name'] = [paths/to/dcm]
    """
    from clinica.utils.stream import cprint
    for ses in dict:
        cprint(ses)
        for dType in dict[ses]:
            cprint('\t' + str(dType))
            for priority in dict[ses][dType]:
                cprint('\t\t' + str(priority))
                for name in dict[ses][dType][priority]:
                    cprint('\t\t\t' + str(name))
                    for path in dict[ses][dType][priority][name]:
                        cprint('\t\t\t\t' + str(path))


def collect_conversion_tuples(final_bids, dest_dir, patient):
    """
    Returns a list of tuples that will be the input of the convert function
    If final_bids[session][datatype][priority][name] contains a path, it is added to tuple[0], tuple[1] contains the path where the Nifti will be located (follows the BIDS format)

    Args:
        final_bids: A data structure of the form : final_bids[session][datatype][priority][name]
        dest_dir: Home of the BIDS directory
        patient: Instantiation of the Patient class

    Returns:
        A list of tuples [(path/to/dicom, path/to/nifti), ...]
    """
    import os
    sol = []
    for ses in final_bids:
        for dType in final_bids[ses]:
            for priority in final_bids[ses][dType]:
                for name in final_bids[ses][dType][priority]:
                    if final_bids[ses][dType][priority][name] != []:
                        path_source = final_bids[ses][dType][priority][name][0]
                        path_dest = os.path.join(dest_dir, patient.get_name(), ses, dType, name)
                        sol.append((path_source, path_dest))
    return sol


def supress_stdout(func):
    '''
    Wrapper, makes a function non-verbose

    Args:
        func: function to be silenced
    '''
    import os
    import contextlib

    def wrapper(*a, **ka):
        with open(os.devnull, 'w') as devnull:
            with contextlib.redirect_stdout(devnull):
                func(*a, **ka)

    return wrapper


def convert_dcm_to_nii(single_tuple):
    """
    :param single_tuple: tuple where tuple[0] is the path to the data,
    and tuple[1] the path to the coverted data
    :return:
    """
    import subprocess
    import os
    from clinica.utils.stream import cprint

    filename = os.path.basename(single_tuple[1])
    path_dest = os.path.dirname(single_tuple[1])
    create_folder(path_dest)
    command = 'dcm2niix -b y -z y -o ' + path_dest + ' -f ' + filename + ' ' + single_tuple[0]
    cprint(' Converting ' + os.path.basename(filename).replace('_', ' '))
    subprocess.run(command,
                   shell=True,
                   stdout=subprocess.DEVNULL,
                   stderr=subprocess.DEVNULL)


def convert(list_tuples):
    '''
    Converts a list of tuples = [(path/to/dicom, path/to/nifti), ...]

    Args:
        list_tuples: list of tuples, tuple[0] contains a path to a Dicom file, tuple[1] contains the path where the Nifti file needs to be created.
    '''
    from multiprocessing import Pool, cpu_count

    p = Pool(max(cpu_count() - 1, 1))
    p.map(convert_dcm_to_nii, list_tuples)
